import re
import os
import pandas as pd
import numpy as np
import scipy.sparse as sp


def convert_sp(input_dir):
	for filename in os.listdir(input_dir):
		if not os.path.exists(input_dir+'/'+filename+'/overlap_matrix.csv'):continue
		ama=input_dir+'/'+filename+'/all_strains_re.csv'
		oma=input_dir+'/'+filename+'/overlap_matrix.csv'
		ma=pd.read_csv(ama)
		ma=ma.values[:,:]
		mo=pd.read_csv(oma)
		mo=mo.values[:,:]
		ma=sp.csr_matrix(ma)
		mo=sp.csr_matrix(mo)
		sp.save_npz(input_dir+'/'+filename+'/all_strains_re.npz',ma)
		sp.save_npz(input_dir+'/'+filename+'/overlap_matrix.npz',mo)
		di=input_dir+'/'+filename+'/all_strain.csv'
		os.system('rm '+ama+' '+oma+' '+di)



from pandas import read_csv
import numpy as np
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import RepeatedKFold
from sklearn.linear_model import ElasticNet, ElasticNetCV
from sklearn.model_selection import ShuffleSplit
from sklearn import metrics
from scipy.stats import pearsonr
from itertools import chain
import pickle
import re
import scipy.sparse as sp

def lasso_mpm(alphas, mse_path):
	mse_mean = np.mean(mse_path, axis=1)
	mse_std = np.std(mse_path, axis=1)
	mse_min_idx = np.argmin(mse_mean)
	mse_min = mse_mean[mse_min_idx]
	mse_min_std = mse_std[mse_min_idx]
	mse_min_std_min = mse_min - mse_min_std
	mse_min_std_max = mse_min + mse_min_std
	
	mse_mpm_idx = mse_min_idx
	for i in range(mse_min_idx-1, -1, -1):
		if (mse_mean[i]>=mse_min_std_min) and (mse_mean[i]<=mse_min_std_max):
			mse_mpm_idx = i
	alpha_mpm = alphas[mse_mpm_idx]
	mse_mean_mpm = mse_mean[mse_mpm_idx]
	mse_std_mpm = mse_std[mse_mpm_idx]

	return alpha_mpm, mse_mean_mpm, mse_std_mpm

def stat_cov(ix,iy):
	cov=0
	total_kmr=np.count_nonzero(ix)
	ic=ix*iy
	ic[ic==1]=0
	valid_kmr=np.count_nonzero(ic)
	if total_kmr==0:
		cov=0
	else:
		cov=float(valid_kmr/total_kmr)
	return [cov,valid_kmr,total_kmr]
def cal_cov_all(ix,iy):
	cov=[]
	for i in range(len(ix[0])):
		arr=stat_cov(ix[:,i],iy)
		cov.append(arr[0])
	return cov
	
def reject_outliers(data,x, n = 3):
	mean=np.mean(data)
	sigma=np.std(data)
	remove_idx=np.where(abs(data-mean)>n*sigma)
	new_y=np.delete(data, remove_idx)
	new_x=np.delete(x, remove_idx, axis=0)
	'''
	d = np.abs(data - np.median(data))
	mdev = np.median(d)
	remove_idx=np.where(abs(data-np.median(data))>m*mdev)
	new_y=np.delete(data, remove_idx)
	new_x=np.delete(x, remove_idx,axis=0)
	'''
	#print(new_y.shape,new_x.shape)
	#exit()
	#print(remove_idx)
	#exit()
	#s = d/mdev if mdev else 0.
	#return data[s<m]
	return new_y,new_x

def merge_x(dX,all_id,cls_info,dominat,dx):
	c=1
	d={}
	X=dX
	#X=np.array(X)
	for i in all_id:
		d[i]=c
		c+=1
	f=open(cls_info,'r')
	new_x=dx
	while True:
		line=f.readline().strip()
		if not line:break
		ele=line.split('\t')
		strains=re.split(',',ele[-1])
		if not int(d[ele[1]])==int(dominat):continue
		for s in strains:
			if int(d[s])==int(dominat):continue
			new_x=new_x+X[:,int(d[s])-1]
	new_x[new_x>1]=1
	return new_x

def get_remainc(dominat,used_kmer,pXt_tem,py,strain_remainc):
	npXt=2*used_kmer+pXt_tem
	npXt[npXt>1]=0
	for i in range(len(npXt)):
		if i==dominat:continue
		all_k=np.sum(npXt[i])
		tem_c=npXt[i]*py
		tem_c[tem_c==1]=0
		tem_c[tem_c>1]=1
		check=np.sum(tem_c)
		if all_k==0:
			strain_remainc[i]=0
		else:
			strain_remainc[i]=check/all_k
	return strain_remainc

def get_avg_depth(dominat,pX,py):
	doarr=pX[:,dominat]*py
	doarr[doarr==1]=0
	doarr_noz=doarr[doarr!=0]
	f25=np.percentile(doarr_noz,25,interpolation='nearest')
	f75=np.percentile(doarr_noz,75,interpolation='nearest')
	doarr_noz[doarr_noz<f25]=0
	doarr_noz[doarr_noz>f75]=0
	doarr_final=doarr_noz[doarr_noz!=0]
	avg_depth=np.mean(doarr_final)
	return avg_depth
def get_candidate_arr(ix,iy):
	res={}
	c=0
	for n in ix:
		tem_c=ix[c]*iy
		tem_c[tem_c==1]=0
		tem_c[tem_c>1]=1
		check=np.sum(tem_c)
		res[c]=check
		c+=1
	hc=sorted(res.items(),key=lambda d:d[1],reverse=True)
	candidate=hc[0][0]
	return candidate,hc[0][1]

def optimize_dominat_y(ix,iy):
	c=0
	res=[]
	#res_std=[]
	#res_cv=[]
	#exit()
	for i in range(ix.shape[1]):
		da=ix[:,c]*iy
		#print(len(ix[:,c]),len(iy))
		#print(ix[:,c])
		da_noz=da[da!=0]
		#exit()
		if np.sum(da_noz)==0 or len(da_noz)<1:
			res.append(0)
			#res_std.append(0)
			#res_cv.append(0)
		else:
			f25=np.percentile(da_noz,5,interpolation='nearest')
			f75=np.percentile(da_noz,95,interpolation='nearest')
			tem_iy=np.copy(iy)
			tem_iy[tem_iy<f25]=0
			tem_iy[tem_iy>f75]=0
			res.append(np.dot(ix[:,c].T,tem_iy))
			'''
			da_noz[da_noz<f25]=0
			da_noz[da_noz>f75]=0
			da_end=da_noz[da_noz!=0]
			me=np.mean(da_end)
			std=np.std(da_end)
			cv=std/me
			res_std.append(std)
			res_cv.append(cv)
			'''
		c+=1
	#exit()
	res=np.array(res)
	dominat=np.where(res==np.max(res))[0][0]
	#print(res)
	#exit()
	return dominat

def detect_strains(input_csv,input_y,ids,ksize,npp25,npp75,npp_out,cls_cov,omatrix,all_cls,l2,msn):

	#data_frame1=read_csv(input_csv)
	new_als=[]
	for a in all_cls:
		new_als.append(int(a-1))
	#data_frame2=read_csv(omatrix,usecols=new_als)
	#dX=data_frame2.values[:,:]
	#dX=np.array(dX)

	#all_id=pickle.load(open(raw_id,"rb"))
	#X=data_frame1.values[:,:]
	
	#om=data_frame2.values[:,:]
	omx=sp.load_npz(omatrix)
	om=omx.A
	om=om[:,new_als]
	#y=data_frame2.values
	#y=list(chain(*y))
	ln=np.sum(om,axis=1)
	ln[ln>1]=0

	#pX=np.array(X)
	X=sp.load_npz(input_csv)
	pX=X.A

	#py=np.array(y)
	py=input_y
	py_u=input_y*ln
	'''
	i=0
	pX=[]
	py=[]
	for v in opy:
		if v<=npp25 or v>=npp75:
			i+=1
			continue
		else:
			pX.append(opX[i])
			py.append(v)
			i+=1
	pX=np.array(pX)
	py=np.array(py)
	'''
	


	sid=pickle.load(open(ids, "rb"))
	 
	cutoff=msn*ksize
	#exit()
	def Pre_Scan(pX,py,sid,cutoff,cls_cov,py_u,l2):
		strain_cov={} # Strain -> [Coverage,covered_kmr,total_kmr]
		strain_val={} # Strain -> Valid kmr
		strain_remainc={}
		final_src={}
		res_std=[]
		res_cv=[]
		mannual_depth={} # Strain -> Mannual Depth
		out_columns=[]
		out_strain=[]
		pXt=pX.T
		cov_arr=cal_cov_all(pX,py)
		cov_arr=np.array(cov_arr)
		#print(cov_arr)
		#print(sid)
		#exit()
		dominat_avg_depth=0
		default_cov=0.7
		'''
		if cls_cov>=0.9:
			default_cov=0.9
		'''
		if np.max(cov_arr)>default_cov: # Filter strains that cov < 0.7
			cov_arr[cov_arr<=default_cov]=0
			cov_arr[cov_arr>default_cov]=1
			pXt_tem=((pXt.T)*cov_arr).T
		else:
			pXt_tem=pXt
			if np.max(cov_arr)<0.01:
				l2=2
		
		#print(ld)
		#exit()
		#py_tem[py_tem<0]=0
		'''
		tem_check_arr=pXt_tem*py
		o=open('Tem_check.txt','w+')
		c=0
		for t in tem_check_arr:
			o.write(sid[c]+'\t'+str(np.mean(t))+'\t'+str(np.median(t))+'\n')
			c+=1
		exit()
		'''
		if l2==2:
			dominat=np.where(cov_arr==np.max(cov_arr))[0][0]
			if np.sum(py_u)>0:
				dominat_avg_depth=get_avg_depth(dominat,pX,py_u)
			else:
				dominat_avg_depth=get_avg_depth(dominat,pX,py)
		else:
			if np.sum(py_u)>0:
				dominat=optimize_dominat_y(pX,py_u)
				dominat_avg_depth=get_avg_depth(dominat,pX,py_u)
			else:
				dominat=optimize_dominat_y(pX,py)
				dominat_avg_depth=get_avg_depth(dominat,pX,py)
			#dominat_avg_depth=get_avg_depth(dominat,pX,py)
		##dominat_global=int(data_frame1.columns[dominat])
		#new_x=merge_x(dX,all_id,cls_info,dominat_global,pXt[dominat])

		out_columns.append(dominat)
		out_strain.append(sid[dominat])
		strain_cov[sid[dominat]]=stat_cov(pX[:,dominat],py)
		strain_val[sid[dominat]]=strain_cov[sid[dominat]][1]
		strain_remainc[sid[dominat]]=stat_cov(pX[:,dominat],py)[0]
		final_src[sid[dominat]]=stat_cov(pX[:,dominat],py)[0]
		
		# Now we will start iterative pre-scan process to find multiple strains.
		max_iter=15 # The maximum iterate times
		used_kmer=pXt[dominat]
		#used_kmer=new_x
		strain_remainc=get_remainc(dominat,used_kmer,pXt_tem,py_u,strain_remainc)

		for i in range(max_iter):
			#if remain_kmer<cutoff:break
			npXt=2*used_kmer+pXt_tem
			npXt[npXt>1]=0
			#npXt[npXt==3]=0
			#candidate_arr=np.dot(npXt,py)
			#if np.max(candidate_arr)<2*cutoff:break
			#candidate=np.where(candidate_arr==np.max(candidate_arr))[0][0]
			if np.sum(py_u)>0:
				candidate,check=get_candidate_arr(npXt,py_u)
			else:
				candidate,check=get_candidate_arr(npXt,py)
			#print(candidate,check)
			#tem_c=npXt[candidate]*py
			#tem_c[tem_c==1]=0
			#tem_c[tem_c>1]=1
			#check1=np.sum(tem_c)
			#print(check1)
			'''
			tem_c=tem_c*py_u # remove union k-mers from other cluster
			tem_c[tem_c<0]=0
			'''
			#check=np.sum(tem_c)


			#print(sid[candidate],check)
			#exit()
			if check>=cutoff:
				if strain_remainc[candidate]>0.2:
					out_columns.append(candidate)
					out_strain.append(sid[candidate])
					strain_cov[sid[candidate]]=stat_cov(pX[:,candidate],py)
					strain_val[sid[candidate]]=check
					final_src[sid[candidate]]=strain_remainc[candidate]

				#candidate_global=int(data_frame1.columns[candidate])
				#new_x=merge_x(dX,all_id,cls_info,candidate_global,pXt[candidate]) 
				used_kmer=used_kmer+pXt[candidate]
				#used_kmer=used_kmer+new_x
				used_kmer[used_kmer>1]=1
			else:
				break
		return out_columns,out_strain,strain_cov,strain_val,final_src,dominat_avg_depth

	out_columns,out_strains,strain_cov,strain_val,final_src,dominat_avg_depth=Pre_Scan(pX,py,sid,cutoff,cls_cov,py_u,l2)
	#print(out_strains,res_std,res_cv)
	#print(out_strains)
	#exit()
	if len(out_columns)==1:
		res=dict(zip(out_strains,[1]))
		res2=dict(zip(out_strains,[dominat_avg_depth]))
		return res,res2,strain_cov,strain_val,final_src

	
	#out_strains=sid
	oX=pX[:,out_columns]
	#X=pX
	oy=py
	#print(out_strains)
	#print(npp25,npp75)
	#X=oX
	#y=py
	#tem_check_arr=X*y
	#for t in tem_check_arr:


	#y,X=reject_outliers(oy,oX)
	
	
	i=0
	X=[]
	y=[]
	for v in oy:
		#if v<=1:
		if v<npp25 or v>npp75 or v>npp_out:
			i+=1
			continue
		else:
			X.append(oX[i])
			y.append(v)
			i+=1
	X=np.array(X)
	y=np.array(y)
	'''
	tem_check_arr=(X.T)*y
	o=open('Tem_check.txt','w+')
	c=0
	for t in tem_check_arr:
		o.write(out_strains[c]+'\t'+str(np.mean(t))+'\t'+str(np.median(t))+'\n')	
		c+=1
	exit()
	'''



	#print(X,y)
	#exit()

	print('Pre-scan finished, now we will start ElasticNet fitting...')

	CV_NITER = 20
	NALPHA = 50
	MAX_NITER = 5000
	TEST_SIZE = 0.5
	cv = ShuffleSplit(n_splits=CV_NITER, test_size=TEST_SIZE, random_state=0)


	lasso_cv = ElasticNetCV(eps=0.001, n_alphas=NALPHA,fit_intercept=False, normalize=False,precompute='auto', max_iter=MAX_NITER,tol=0.0001, copy_X=True, cv=cv, verbose=False,n_jobs=1, positive=True, random_state=0,selection='cyclic')

	lasso_cv.fit(X, y)

	alpha, mse_ave, mse_std = lasso_mpm(lasso_cv.alphas_, lasso_cv.mse_path_)

	#MAX_NITER = 10000

	#print(alpha)
	# Main Part

	lasso = ElasticNet(alpha=alpha, fit_intercept=False, normalize=False,precompute=False, copy_X=True, max_iter=MAX_NITER,tol=0.0001, warm_start=False, positive=True,random_state=0, selection='cyclic')

	#lasso=Lasso(alpha=alpha)

	lasso.fit(X, y)
	lasso_coef = np.atleast_1d(lasso.coef_)
	#fy=lasso.predict(X)
	#print ("RMSE:", np.sqrt(metrics.mean_squared_error(y, fy)))
	#exit()

	#print(lasso_coef)
	#print(out_strains)
	#exit()

	if not np.sum(lasso_coef)==0:
		coef_norm = lasso_coef / np.sum(lasso_coef)
		res=dict(zip(out_strains,list(coef_norm)))
		res2=dict(zip(out_strains,list(lasso_coef)))
	else:
		res={}
		res2={}

	#print(coef_norm)
	#res=dict(zip(out_strains,list(coef_norm)))
	#res2=dict(zip(out_strains,list(lasso_coef)))
	return res,res2,strain_cov,strain_val,final_src

	#import pickle
	#sid=pickle.load(open("../StrainVote_DB_Lasso_Test/Kmer_Sets/Kmer_Sets/C6/All_Kmer/id2strain_re.pkl", "rb")) 
	#print(sid)
	#print(zip(list(lasso_coef),sid))

	#y_pred = X.dot(coef_norm)
	#r,pval = pearsonr(y, y_pred)
	#print(r,pval)


#print(lasso_coef)






import scipy.stats as st
from collections import defaultdict
import numpy as np
import os
import time


def identify_cluster(fq_path, db_dir):
    start=time.time()

    # tree structure
    class tree_node():
        def __init__(self):
            self.seq = -1
            self.leaves = []
            self.child = []
            self.father = None
            self.access = 0 # match

            self.total_num = 0
            self.covered_num = 0
            self.strain = ""

        def get_child(self, T1, T2):
            self.child.append(T1)
            self.child.append(T2)
            T1.father = self
            T2.father = self

    # read tree
    nodes = []
    leaves = []
    f = open(db_dir+"/tree_structure.txt", "r")
    lines = f.readlines()
    for line in lines:
        temp = line.rstrip().split("\t")
        T = tree_node()
        T.seq = int(temp[0])
        if(temp[1] == "N"):
            T0 = T
        else:
            T.father = int(temp[1]) # num
        nodes.append(T)
        if(temp[2] != "N"):
            temp1 = temp[2].split(" ")
            T.child = [int(temp1[0]), int(temp1[1])]
        else:
            leaves.append(T)
        temp1 = temp[3].split(" ")
        del temp1[-1]
        for i in temp1:
            T.leaves.append(int(i))
        if(len(temp) == 5):
            T.strain = temp[4]

    # get children and father
    for i in nodes:
        if(i.father != None):
            for j in nodes:
                if(j.seq == i.father):
                    i.father = j
        if(i.child != []):
            temp = []
            for j in i.child:
                for k in nodes:
                    if(k.seq == j):
                        temp.append(k)
            i.child = temp

    # zero nodes
    length_z = {}
    f = open(db_dir+"/nodes_length.txt", "r")
    lines = f.readlines()
    for line in lines:
        d = line.rstrip().split("\t")
        length_z[int(d[0])] = int(d[1])
    nodes_spec = []
    nodes_s = []
    for i in nodes:
        if(i == T0):
            continue
        if(length_z[i.seq] < 300):
            nodes_s.append(i)
        if(length_z[i.seq] < 300 and length_z[find_bro(i).seq] < 300):
            nodes_spec.append((i, find_bro(i)))

    # k-mer count
    if(type(fq_path) == str):
        x = fq_path
    else:
        s = " "
        x = s.join(fq_path)
    dir_jf = os.path.split(os.path.abspath(__file__))[0]+'/jellyfish-linux'
    path = "temp.fa"
    line = dir_jf + " count -m 31 -s 100M -t 8 --if " + db_dir+"/kmer.fa -o temp.jf " + x
    os.system(line)
    line = dir_jf+" dump -c temp.jf > temp.fa"
    os.system(line)
    line = "rm temp.jf"
    os.system(line)
    match_results = {}
    f = open(path, "r")
    lines = f.readlines()
    for line in lines:
        temp = line.rstrip().split(" ")
        match_results[temp[0]] = int(temp[1])
    end = time.time()
    print('- The total running time of jellyfish match is ',str(end-start),' s\n')
    os.system("rm temp.fa")

    # overlap nodes
    overlap_nodes = []
    f = open(db_dir + "/overlap_nodes", "r")
    lines = f.readlines()
    for line in lines:
        overlap_nodes.append(int(line.rstrip()))

    # identification
    start = time.time()
    pending = [(T0.child[0], T0.child[1])]
    results = []
    abundance = {}
    cov = {}
    length = {}
    label = {}  # whether need correction
    for i in nodes:
        label[i.seq] = 0
    while(len(pending) != 0):
        for data in pending:
            # < 300 nodes, special processing
            if(data in nodes_spec):
                print(data[0].seq, data[1].seq)
                for i in data:
                    pending.append((i.child[0], i.child[1]))
                    abundance[i.seq] = abundance[i.father.seq]
                    cov[i.seq] = 0
                    length[i.seq] = 0
                    i.access = 2
                pending.remove(data)
                continue

            zero_label = 0
            print("--------------------------------------------------")
            for i in data:
                if(i in nodes_s):
                    print("node %d default"%i.seq)
                    zero_label = 1
                    abundance[i.seq] = 0
                    cov[i.seq] = 0
                    length[i.seq] = 0
                    X = i
                    continue
                print("node %d"%i.seq)
                Y = i   # not zero
                f = open(db_dir+"/nodes_kmer/"+str(i.seq), "r")
                lines = f.readlines()
                f.close()
                profile = []
                uncovered = 0
                if(i.seq not in overlap_nodes or len(results) == 0):
                    line = lines[0].rstrip().split(" ")
                    for x in line:
                        if(x in match_results):
                            if(match_results[x]>0): # dont ignore 1
                                profile.append(match_results[x])
                            else:
                                uncovered += 1
                else:
                    temp_match = {}
                    line = lines[0].rstrip().split(" ")

                    # calculate unique kmer quantity
                    length_temp = len(line)
                    delete = set([])
                    Overlap = defaultdict(list)
                    for j in results:
                        path = db_dir+"/overlap/"+str(i.seq)+"_"+str(j.seq)
                        if(os.path.exists(path) == False):
                            continue
                        f = open(path, "r")
                        lines = f.readlines()
                        if(len(lines)>0):
                            Overlap[j.seq] = list(map(int, lines[0].rstrip().split(" ")))
                            delete = set(Overlap[j.seq]) | delete
                    if(length_temp - len(delete) >= 1000):
                        remain = set(list(range(0, length_temp))) - delete
                        remain = list(remain)
                        for k in remain:
                            if(line[k] in match_results):
                                if(match_results[line[k]]>0):
                                    profile.append(match_results[line[k]])
                                else:
                                    uncovered += 1
                    else:
                        j = 0
                        label[i.seq] = 1
                        for x in line:
                            if(x in match_results):
                                temp_match[j] = match_results[x]
                            j += 1
                        for j in results:
                            path = db_dir+"/overlap/"+str(i.seq)+"_"+str(j.seq)
                            if(os.path.exists(path) == False):
                                continue
                            temp_match1 = {}
                            if(j.seq in Overlap):
                                for k in Overlap[j.seq]:
                                    x = int(k)
                                    if(temp_match[x]>0):
                                        temp_match1[x] = temp_match[x]
                                for k in results:
                                    print(k.seq)
                                print(j.seq, abundance[j.seq])
                                sample = np.random.poisson(abundance[j.seq], size=len(temp_match1))
                                sample.sort()
                                Tuple = sorted(temp_match1.items(), key = lambda kv:(kv[1], kv[0]))
                                for j in range(0, len(sample)):
                                    temp_match[Tuple[j][0]] = Tuple[j][1] - sample[j]
                        for j in temp_match:
                            if(temp_match[j]>0):
                                profile.append(temp_match[j])
                            else:
                                uncovered += 1

                length[i.seq] = uncovered + len(profile)
                if(length[i.seq] == 0):
                    cov[i.seq] = 0
                else:
                    cov[i.seq] = len(profile)/length[i.seq]
                cov_cutoff = 0.5
                if(cov[i.seq] <= cov_cutoff):   # temporary abundance
                    abundance[i.seq] = 0
                else:
                    abundance[i.seq] = np.mean(profile)
                print("%f | %f"%(abundance[i.seq], cov[i.seq]))

            # binomial test
            if(zero_label == 1):
                abundance[X.seq] = abundance[X.father.seq] - abundance[Y.seq]
            ab_temp = {}
            for i in range(0, 2):
                ab_temp[data[i]] = round(abundance[data[i].seq])
            if(list(ab_temp.values()) == [0, 0]):
                pending.remove(data)
                continue
            Tuple = sorted(ab_temp.items(), key = lambda kv:(kv[1], kv[0]))
            (a, b, x, y) = (Tuple[1][0], Tuple[0][0], Tuple[1][1], Tuple[0][1])
            ret = 1 - st.binom.sf(max([x, y]), x+y, 0.995)
            get_ab = []
            if(ret < 0.05):
                for i in (a, b):
                    get_ab.append(i)
                    i.access = 1
                    correct_access(i)
                    if(i not in leaves):
                        pending.append((i.child[0], i.child[1]))
                    else:
                        results.append(i)
            else:
                a.access = 1
                get_ab.append(a)
                correct_access(a)
                if(a not in leaves):
                    pending.append((a.child[0], a.child[1]))
                else:
                    results.append(a)
            pending.remove(data)

            # get true abundance
            for i in get_ab:
                if(label[i.seq] == 1):
                    path_t = []
                    get_tree_path(i, path_t)
                    Sum_l = 0
                    for j in path_t:
                        Sum_l += length[j.seq]*cov[j.seq]
                    #elif(label_r[i.seq] == 1 and label_r[i.seq] == 0)
                    if(Sum_l == 0):
                        continue
                    ratio_cov = []
                    ab_t = []
                    for j in path_t:
                        ratio_cov.append(cov[j.seq]*length[j.seq]/Sum_l)
                        ab_t.append(abundance[j.seq])
                    abundance_t = sum([a*b for a,b in zip(ab_t,ratio_cov)])
                    if(find_bro(i).access == 1 and label[find_bro(i).seq] == 0):
                        abundance[i.seq] = abundance_t - abundance[find_bro(i).seq]
                    elif(find_bro(i).access == 1 and label[find_bro(i).seq] == 1):
                        abundance[i.seq] = abundance[i.seq]/(abundance[i.seq]+abundance[find_bro(i).seq]) * abundance_t
                        abundance[find_bro(i).seq] = abundance_t - abundance[i.seq]
                        get_ab.remove(find_bro(i))
                    else:
                        abundance[i.seq] = abundance_t

    total_ab = 0
    for i in results:
        path_t = []
        get_uniq_path(i, path_t)
        Sum_l = 0
        Sum_t = 0
        if T0 in path_t:
            path_t.remove(T0)
        for j in path_t:
            Sum_l += length[j.seq]*cov[j.seq]
            Sum_t += length[j.seq]
        ratio_cov = []
        ab_t = []
        i.total_num = Sum_t
        i.covered_num = Sum_l
        for j in path_t:
            ratio_cov.append(cov[j.seq]*length[j.seq]/Sum_l)
            ab_t.append(abundance[j.seq])
        #total_ab += abundance[i.seq]
        abundance_t = sum([a*b for a,b in zip(ab_t,ratio_cov)])
        abundance[i.seq] = abundance_t
        total_ab += abundance[i.seq]

    # output
    res = defaultdict(lambda:{})
    for i in results:
        print(i.seq, abundance[i.seq])
        res[i.seq]['cls_ab'] = abundance[i.seq]
        res[i.seq]['cls_per'] = abundance[i.seq]/total_ab
        res[i.seq]['cls_cov'] = i.covered_num/i.total_num
        res[i.seq]['cls_total_num'] = i.total_num
        res[i.seq]['cls_covered_num'] = i.covered_num
        res[i.seq]['strain'] = 0
        res[i.seq]['s_ab'] = 0

        if(i.strain != ""):
            res[i.seq]['strain'] = i.strain
            res[i.seq]['s_ab'] = abundance[i.seq]
    end = time.time()
    print('- The total running time of identification is ',str(end-start),' s\n')
    return res


def find_bro(T):
    for i in T.father.child:
        if(i!=T):
            return i


def get_tree_path(i, path_t):
    i = i.father
    path_t.append(i)
    if(i.father == None or find_bro(i).access == 1):
        return 1
    else:
        get_tree_path(i, path_t)


def correct_access(T):
    T.access = 1
    if(T.father == None):
        return
    else:
        correct_access(T.father)


def get_uniq_path(i, path_t):
    path_t.append(i)
    if(i.father == None or find_bro(i).access == 1):
        return 1
    get_uniq_path(i.father, path_t)


#fq = ("reads/1.fq", "reads/2.fq")

# real reads
#fq = ("/home/yongxinji2/data/real_reads/LOR054C_S50_R1_001_adapter_removed_1.fastq", "/home/yongxinji2/data/real_reads/LOR054C_S50_R1_001_adapter_removed_2.fastq")
#fq = ("/home/yongxinji2/data/real_reads/LOR034C_S51_R1_001_adapter_removed_1.fastq", "/home/yongxinji2/data/real_reads/LOR034C_S51_R1_001_adapter_removed_2.fastq")
#fq = ("/home/yongxinji2/data/real_reads/LOR295C_S132_R1_001_adapter_removed_1.fastq", "/home/yongxinji2/data/real_reads/LOR295C_S132_R1_001_adapter_removed_2.fastq")
#fq = ("/home/yongxinji2/data/SRR3184328.fq")
#fq=("Sim_Data/GCA_018128405_1.fq","Sim_Data/GCA_018128405_2.fq")
#fq=("Ecoil_Real_Data/ERR260499_1.fastq","Ecoil_Real_Data/ERR260499_2.fastq")
#fq=("Tem_Real/ERR260487_1.fastq","Tem_Real/ERR260487_2.fastq")
#results = identify_cluster(fq, "/home/yongxinji2/worktemp/Tree_database_ecoli")
#print(dict(results))
#for i in results.keys():
#    print(i, results[i])



import re
import os
import sys

def construct_matrix(input_fa):	
	'''
	pwd=os.getcwd()
	dir_work=os.path.split(os.path.abspath(__file__))[0]
	os.chdir(dir_work)
	'''
	path_file='genome_path_tem.txt'
	o=open(path_file,'w+')
	for filename in os.listdir(input_fa):
		o.write(input_fa+'/'+filename+'\n')
	o.close()
	#pwd=os.getcwd()	
	dir_dash=os.path.split(os.path.abspath(__file__))[0]+'/dashing_s128'
	#print(dir_dash+' dist -p10 -k31 -Odistance_matrix.txt -osize_estimates.txt -Q '+path_file+' -F '+path_file)
	#print(pwd)
	#exit()
	os.system(dir_dash+' dist  -p10 -k31 -Odistance_matrix.txt -osize_estimates.txt -Q  '+path_file+' -F '+path_file)
	#exit()
	nn=rebuild_matrix('distance_matrix.txt')
	os.system('rm genome_path_tem.txt size_estimates.txt')
	#os.system('rm genome_path_tem.txt size_estimates.txt')
	return nn
				

def rebuild_matrix(input_m):
	fn=os.path.basename(input_m)
	pre=re.sub('\..*','',fn)
	nn=pre+'_rebuild.txt'
	f=open(input_m,'r')
	o=open(nn,'w+')
	line=f.readline().strip()
	ele=line.split('\t')[1:]
	for e in ele:
		o.write('\t'+e)
	o.write('\n')
	while True:
		line=f.readline().strip()
		if not line:break
		ele=line.split('\t')
		teme=[]
		for e in ele[1:]:
			teme.append(str(float(1)-float(e)))
		o.write(ele[0])
		tem='\t'.join(teme)
		o.write('\t'+tem+'\n')
	return nn


def hcls(input_m,method,cutoff):
	o=open('tem_hier.R','w+')
	o.write('x<-read.table(\"'+input_m+'\", header=T, row.names=1)\nd<-as.dist(as(x,\"matrix\"))\nhc<-hclust(d,method=\"'+method+'\")\nres<-sort(cutree(hc,h='+str(cutoff)+'))\nres')
	o.close()
	os.system('Rscript tem_hier.R > cls_res.txt')
	os.system('rm tem_hier.R')
	f=open('cls_res.txt','r')
	a=[]
	while True:		
		line=f.readline().strip()
		if not line:break
		#if re.search('/',line):
		a.append(line)
	d={} # {'1':{'../G1.fna':'','../G2/fna',..},'2':...}
	dmap={}
	c=0
	for l in a[::-1]:
		c+=1
		if not c%2==0:
			ele=l.split()
			if len(ele)==1:
				if l not in d:
					d[l]={}
					dmap[l]={}
				name=l
			else:
				for e in ele:
					if e not in d:
						d[e]={}
						dmap[e]={}
				name=ele
		else:
			ele=l.split()
			if len(ele)==1:
				d[name][l]=''
				pre=os.path.split(l)[1]
				pre=re.split('\.',pre)[0]
				dmap[name][pre]=''
			else:
				i=0
				for e in ele:
					d[name[i]][e]=''
					pre=os.path.split(e)[1]
					pre=re.split('\.',pre)[0]
					dmap[name[i]][pre]=''
					i+=1
	#print(int(100-float(cutoff)*100))
	o2=open('hclsMap_'+str(int(100-float(cutoff)*100))+'.txt','w+')
	for e in dmap:
		tems=','.join(dmap[e])
		o2.write(str(e)+'\t'+str(len(dmap[e]))+'\t'+tems+'\n')
	os.system('rm cls_res.txt')
	return d



	





import re
import os
import sys
import argparse
sys.path.append('library')
from collections import defaultdict
from library import identify,Vote_Strain_L2_Lasso_new_sp
import pickle




__author__="Liao Herui, Ji Yongxin - PhD of City University of HongKong"
usage="StrainScan - A kmer-based strain-level identification tool."

def initial_para(para,value):
	if not para:
		para=value
	return para

def build_dir(in_dir):
	if not os.path.exists(in_dir):
		os.makedirs(in_dir)

def get_overlap_kmr(db_dir,apcls,pcls):
	overlap_kmr=defaultdict(lambda:{})
	carr=[]
	for c in apcls:
		carr.append(c)
	all_k={}
	for c in carr:
		#if c not in pcls:continue
		all_k[c]=pickle.load(open(db_dir+'/Kmer_Sets_L1/Kmer_Sets/'+c+'/all_kid.pkl','rb'))
	overlap_kmr={}
	for c in apcls:
		if c not in pcls:continue
		cud=all_k[c]
		overlap_kmr[c]={}
		for c2 in apcls:
			if c==c2:continue
			pov1=cud.keys() & all_k[c2].keys()
			pov=dict.fromkeys(pov1,'')
			overlap_kmr[c]=dict(overlap_kmr[c],**pov)

	return overlap_kmr

def main():
	pwd=os.getcwd()
	# Get para
	parser=argparse.ArgumentParser(prog='StrainScan.py',description=usage)
	parser.add_argument('-i','--input_fastq',dest='input_fq',type=str,required=True,help="The dir of input fastq data --- Required")
	parser.add_argument('-d','--database_dir',dest='db_dir',type=str,required=True,help="The dir of your database --- Required")
	parser.add_argument('-o','--output_dir',dest='out_dir',type=str,help='Output dir (default: current dir/StrainVote_Result)')
	parser.add_argument('-k','--kmer_size',dest='ksize',type=str,help='The size of kmer, should be odd number. (default: k=31)')
	parser.add_argument('-l','--low_dep',dest='ldep',type=str,help='This parameter can be set to \"1\" if the sequencing depth of input data is very low (e.g. < 10x). For super low depth ( < 1x ), you can use \"-l 2\"  (default: -l 0)')
	parser.add_argument('-s','--minimum_snv_num',dest='msn',type=str,help='The minimum number of SNV at Layer-2 identification. (default: k=40)')

	args=parser.parse_args()
	fq_dir=args.input_fq
	db_dir=args.db_dir
	out_dir=args.out_dir
	ksize=args.ksize
	ksize=initial_para(ksize,31)
	ldep=args.ldep
	msn=args.msn
	if not ldep:
		ldep=0
	else:
		ldep=int(ldep)
	if not msn:
		msn=40
	else:
		msn=int(msn)
	
	out_dir=initial_para(out_dir,pwd+'/StrainScan_Result')
	if not re.search('/',out_dir):
		out_dir=pwd+'/'+out_dir
	build_dir(out_dir)

	# Step1 -> Vote possible clusters using Krakenuniq
	#pcls,uniq_strain,apcls=Vote_Cls_KK.vote_cls(fq_dir,db_dir,out_dir)
	#overlap_kmr=get_overlap_kmr(db_dir,apcls,pcls)
	# Step2 -> Vote Cls
	in_fq=(fq_dir)
	l2=0
	#cls_dict=identify_cluster_u.identify_cluster(in_fq,'/home/yongxinji2/worktemp/Tree_database')
	if ldep==0:
		cls_dict=identify.identify_cluster(in_fq,db_dir+'/Tree_database',[0.1,0.4,1])
		if len(cls_dict)==0:
			cls_dict=identify.identify_cluster(in_fq,db_dir+'/Tree_database',[0.05,0.05,1])
			l2=1
		if len(cls_dict)==0:
			print('Warning: No clusters can be detected!')
			exit()
	elif ldep==1:
		cls_dict=identify.identify_cluster(in_fq,db_dir+'/Tree_database',[0.01,0.05,1])
		l2=1
	elif ldep==2:
		cls_dict=identify.identify_cluster(in_fq,db_dir+'/Tree_database',[0.005,0.01,1])
		l2=1

	#cls_dict.update(uniq_strain)
	if len(cls_dict)==0:
		print('Warning: No clusters can be detected!')
		exit()
	# Step3 -> Vote Strains inside Cls
	Vote_Strain_L2_Lasso_new_sp.vote_strain_L2_batch(fq_dir,db_dir,out_dir,ksize,dict(cls_dict),l2,msn)
	

if __name__=='__main__':
	sys.exit(main())



import re
import os
from collections import defaultdict
from Bio import SeqIO
import seqpy
#from tqdm import tqdm


encoding_map={'A':0,'G':1,'C':2,'T':3}
decoding_lst=['A','G','C','T']

def encode(k):
	code=0
	for ch in k:
		code*=4
		code+=encoding_map[ch]
	return code, len(k)
def decode(enc):
	code, length=enc
	ret=''
	for _ in range(length):
		index=code & 3
		code>>=2
		ret=decoding_lst[index]+ret
	return ret

def special_match(strg,search=re.compile(r'[^ATGC]').search):
	return not bool(search(strg))

def get_pre(file_dir):
	name=os.path.split(file_dir)[1]
	pre=re.split('\.',name)[0]
	return pre

def build_kmer_dict(d,k):
	#dlabel={}
	print('::Build kmer dict for all clusters')
	dlabel=defaultdict(lambda:{})
	c=1
	#dmap={}
	for l in d:
		print('::Process - ',c,'/',len(d))
		single=0
		if len(d[l])==1:
			single=1
		for g in d[l]:
			if single==1:
				pre=get_pre(g)
				#dmap[c]=pre
				#dmap[pre]=c
			else:
				#dmap[c]='C'+str(l)
				#dmap['C'+str(l)]=c
				pre='C'+str(l)
			seq_dict = {rec.id : rec.seq for rec in SeqIO.parse(g, "fasta")}
			#exit()
			for cl in seq_dict:
				seq=str(seq_dict[cl])
				for i in range(len(seq)-k+1):
					if not special_match(seq[i:i+k]):continue
					kmrb=encode(seq[i:i+k])
					rev_kmrb=encode(seqpy.revcomp(seq[i:i+k]))
					#if not special_match(seq[i:i+k]):continue
					dlabel[kmrb][c]=''
					dlabel[rev_kmrb][c]=''
		c+=1
	return dlabel

				
def unique_kmer_out(d,k,dlabel,out_d,base_dir,if_base):
	print('::Scan unique kmer and output')
	count=1
	for cls in d:
		print('::Process - ',count,'/',len(d))
		count+=1
		single=0
		uk_count=0
		resd={} # The dict used to record final unique kmer of each cluster
		single_cls=0
		if len(d[cls])==1:
			single_cls=1
			if not if_base=='Y':
				o=open(out_d+'/C'+str(cls)+'.fasta','w+')
				pre='C'+str(cls)
		else:
			o=open(out_d+'/C'+str(cls)+'.fasta','w+')
			pre='C'+str(cls)
		for g in d[cls]:
			if single_cls==1 and if_base=='Y':
				pre=get_pre(g)
				o=open(base_dir+'/'+pre+'.fasta','w+')
		
			
			seq_dict = {rec.id : rec.seq for rec in SeqIO.parse(g, "fasta")}
			
			for cl in seq_dict:
				seq=str(seq_dict[cl])
				for i in range(len(seq)-k+1):
					if not special_match(seq[i:i+k]):continue
					kmrb=encode(seq[i:i+k])
					if len(dlabel[kmrb])==1:
						resd[kmrb]=''
						resd[encode(seqpy.revcomp(seq[i:i+k]))]=''
		for kmrb in resd:
			uk_count+=1
			kmr=decode(kmrb)
			o.write('>'+str(uk_count)+'\n'+str(kmr)+'\n')





def find_unique_kmers(d,out_dir,uni_kmer,k):
	out_d=uni_kmer+'/'+out_dir
	if not os.path.exists(out_d):
		os.makedirs(out_d)
	### Build kmer dict with genome labels
	dlabel=build_kmer_dict(d,k)
	### Scan and output ####
	#unique_kmer_out(d,k,dlabel,out_d,uni_kmer)
	if out_dir=='Cutoff_95':
		unique_kmer_out(d,k,dlabel,out_d,uni_kmer,'Y')
	else:
		unique_kmer_out(d,k,dlabel,out_d,uni_kmer,'N')
		
#print(seqpy.revcomp('ATGC'))
#print(os.getcwd())
	



import re
import os
import numpy as np
from collections import defaultdict
import sys
sys.path.append(os.path.split(os.path.abspath(__file__))[0])
import identify_strains_L2_Enet_Pscan_new_sp
import pickle
import math
import uuid
import multiprocessing


def load_kmer_count(input_kcount):
	f=open(input_kcount,'r')
	dcount_o={}
	dcount_u={}
	total_kcount=0
	while True:
		line=f.readline().strip()
		if not line:break
		ele=line.split()
		dcount_o[ele[0]]=int(ele[1])
		dcount_u[ele[0]]=int(ele[1])
		total_kcount+=int(ele[1])
	return dcount_o,total_kcount


def parse_name(filename):
	out_name=re.split('\.',filename)[0]
	return out_name

def output_func(d,o,if_unique,dcls,total_kcount):
	o.write('ID\tName\tPercent(%)\tKmer_Freq\tCovered_Kmer\tTotal_Kmer\tKmer_Cov\tCls_Size\tStrains\n')
	drank={}
	for s in d:
		drank[s]=d[s][1]
	res=sorted(drank.items(),key=lambda d:d[1],reverse=True)
	temc=1
	for s in res:
		rate=float(d[s[0]][1])/float(total_kcount)
		prop=str('%.2f%%' % (rate * 100))
		if if_unique=='Y':
			o.write(str(temc)+'\t'+s[0]+'\t'+prop+'\t'+str(d[s[0]][1])+'\t'+str(d[s[0]][2])+'\t'+str(d[s[0]][3])+'\t'+str(float(d[s[0]][2])/float(d[s[0]][3]))+'\t1\t'+s[0]+'\n')
		else:
			o.write(str(temc)+'\t'+s[0]+'\t'+prop+'\t'+str(d[s[0]][1])+'\t'+str(d[s[0]][2])+'\t'+str(d[s[0]][3])+'\t'+str(float(d[s[0]][2])/float(d[s[0]][3]))+'\t'+dcls[s[0]][1]+'\t'+dcls[s[0]][2]+'\n')
		temc+=1
		

def load_cls_info(match_dict,db_dir):
	dcls=defaultdict(lambda:{})
	go_dir=db_dir+'/Cluster_Result'
	for c in match_dict:
		f=open(go_dir+'/'+match_dict[c],'r')
		while True:
			line=f.readline().strip()
			if not line:break
			ele=line.split('\t')
			cls='C'+str(ele[0])
			dcls[c][cls]={1:str(ele[1]),2:str(ele[2])}
	return dcls		

def build_dir(idir):
	if not os.path.exists(idir):
		os.makedirs(idir)

def check_L1_res(res):
	print('- Check L1 identification result firstly ...')
	check=1
	for r in res:
		if res[r]['strain']==0:
			check=0
	return check

def extract_kmr_ratio(db_dir,res):
	ok_percent={} # {'1':0.4,'2':0.2,...}
	overlap_kmr={}
	#union_kmr={} # {'1':{union kmer...},'2':...}
	total_depth=0
	# Load kmer dict to a overall dict
	all_k={}
	for r in res:
		all_k[r]=pickle.load(open(db_dir+'/Kmer_Sets_L2/Kmer_Sets/C'+str(r)+'/all_kid.pkl','rb'))
	for r in res:
		#cud=pickle.load(open(db_dir+'/Kmer_Sets/Kmer_Sets/C'+str(r)+'/all_kid.pkl','rb'))
		#if not res[r]['strain']==0:continue
		cud=all_k[r]
		overlap_kmr[r]={}
		for r2 in res:
			if r==r2:continue
			pov1=cud.keys() & all_k[r2].keys()
			pov=dict.fromkeys(pov1,'')
			overlap_kmr[r]=dict(overlap_kmr[r],**pov)
			
		'''
		if len(overlap_kmr)==0:
			overlap_kmr=cud.keys()
		else:
			overlap_kmr=overlap_kmr & cud.keys()
		'''
		#union_kmr[r]={}
		'''
		for r2 in res:
			if r==r2:continue
			cud2=pickle.load(open(db_dir+'/Kmer_Sets/Kmer_Sets/C'+str(r2)+'/all_kid.pkl','rb'))
			union_kmr[r]=dict(union_kmr[r], **cud2)
		'''
		total_depth+=res[r]['cls_ab']
	for r in res:
		ok_percent[r]=res[r]['cls_ab']/total_depth
	#print(overlap_kmr)
	#exit()
	return overlap_kmr,ok_percent

def merge_res(out_dir,res):
	# Merge related results and output final report
	o=open(out_dir+'/final_report.txt','w+')
	o.write('ID\tStrain_Name\tCluster_ID\tRelative_Abundance\tPredicted_Depth (Enet)\tPredicted_Depth (Ab*cls_depth)\tCoverage\tCoverd/Total_kmr\n')
	dab={} # Strain -> relative abundance
	total_depth=0
	dinfo=defaultdict(lambda:{}) # Strain -> {'cid':'C1','pde':'70.21...' }
	for r in res:
		if not res[r]['strain']==0:
			total_depth+=float(res[r]['s_ab'])
			dinfo[res[r]['strain']]['cid']='C'+str(r)
			dinfo[res[r]['strain']]['pde']='NA'
			dinfo[res[r]['strain']]['pda']=float(res[r]['s_ab'])
			dinfo[res[r]['strain']]['cov']=float(res[r]['cls_cov'])
			dinfo[res[r]['strain']]['ct']=str(res[r]['cls_covered_num'])+'/'+str(res[r]['cls_total_num'])
			dinfo[res[r]['strain']]['percent']=float(res[r]['cls_per'])
		else:
			if not os.path.exists(out_dir+'/C'+str(r)+'/StrainVote.report'):continue
			f=open(out_dir+'/C'+str(r)+'/StrainVote.report','r')
			line=f.readline()
			total_depth_pda=0
			total_depth_pde=0
			snum=0
			tem=[]
			while True:
				line=f.readline().strip()
				if not line:break
				ele=line.split('\t')
				#if float(ele[3])<0.02:continue # Threashold for relative abundance
				#if float(ele[6])<0.75:continue # Threashold for coverage
				total_depth_pda+=float(ele[5])
				total_depth_pde+=float(ele[4])
				dinfo[ele[1]]['cid']=ele[2]
				dinfo[ele[1]]['pde']=str(ele[4])
				dinfo[ele[1]]['pda']=float(ele[5])
				dinfo[ele[1]]['cov']=str(ele[6])
				dinfo[ele[1]]['ct']=str(ele[7])
				dinfo[ele[1]]['percent']=float(res[r]['cls_per'])*float(ele[3])
				tem.append(ele[1])
			#print('C'+str(r),tem)
			#exit()
			if len(tem)==1:
				total_depth+=total_depth_pde
				dinfo[tem[0]]['pda']=float(dinfo[tem[0]]['pde'])
				
			else:
				total_depth+=total_depth_pda
	for s in dinfo:
		dab[s]=dinfo[s]['pda']/total_depth
		#dab[s]=dinfo[s]['percent']
	fr=sorted(dab.items(),key=lambda d:d[1],reverse=True)
	c=1
	for r in fr:
		o.write(str(c)+'\t'+r[0]+'\t'+dinfo[r[0]]['cid']+'\t'+str(r[1])+'\t'+str(dinfo[r[0]]['pde'])+'\t'+str(dinfo[r[0]]['pda'])+'\t'+str(dinfo[r[0]]['cov'])+'\t'+dinfo[r[0]]['ct']+'\n')
		c+=1


def recal_depth_cov(res_single,fq_dir,out_dir,db_dir,overlap_kmr):
	if not os.path.exists(out_dir+'/Tem'):
		os.makedirs(out_dir+'/Tem')
	all_kmr={}
	group_kmr=defaultdict(lambda:{})
	for r in res_single:
		fasta=db_dir+'/Kmer_Sets_L2/Kmer_Sets/C'+str(r)+'/all_kid.pkl'
		fa=pickle.load(open(fasta,'rb'))
		all_kmr.update(fa)
		group_kmr[r]=fa
	all_fa=out_dir+'/Tem/tem_raw.fa'
	o=open(all_fa,'w+')
	c=1
	for kmr in all_kmr:
		o.write('>'+str(c)+'\n'+kmr+'\n')
		c+=1
	o.close()
	dir_jf=os.path.split(os.path.abspath(__file__))[0]+'/jellyfish-linux'
	out_jf=out_dir+'/Tem/Tem.jf'
	out_fa=out_dir+'/Tem/Tem.fa'
	os.system(dir_jf+' count -m 31 -s 100M -t 10 --if '+all_fa+' -o '+out_jf+' '+fq_dir)
	os.system(dir_jf+' dump -c '+out_jf+' > '+out_fa)
	dcount={}
	f=open(out_fa,'r')
	while True:
		line=f.readline().strip()
		if not line:break
		ele=line.split()
		dcount[ele[0]]=int(ele[-1])
	res_single_new=defaultdict(lambda:{})
	for r in res_single:
		count_arr=[]
		coverd=0
		total=0
		for kmr in group_kmr[r]:
			if kmr in overlap_kmr[r]:continue
			if kmr not in dcount:continue
			total+=1
			if dcount[kmr]>1:
				count_arr.append(dcount[kmr])
				coverd+=1
		count_arr1=np.array(count_arr)
		f25=np.percentile(count_arr1,25,interpolation='nearest')
		f75=np.percentile(count_arr1,75,interpolation='nearest')
		count_arr1[count_arr1<f25]=0
		count_arr1[count_arr1>f75]=0
		count_arr=count_arr1[count_arr1!=0]
		res_single_new[r]['cls_cov']=float(coverd)/float(total)
		res_single_new[r]['cls_ab']=np.mean(count_arr)
		res_single_new[r]['cls_covered_num']=coverd
		res_single_new[r]['cls_total_num']=total
		res_single_new[r]['strain']=res_single[r]['strain']
		res_single_new[r]['s_ab']=np.mean(count_arr)
		res_single_new[r]['cls_per']=0
	os.system('rm -rf '+out_dir+'/Tem')
	return res_single_new



def generate_single_report(in_dict,out_dir):
	o=open(out_dir+'/final_report.txt','w+')
	o.write('Strain_ID\tStrain_Name\tCluster_ID\tRelative_Abundance_Inside_Cluster\tPredicted_Depth\tCoverage\tCovered/Total_kmr\n')
	res_tem={}
	sc={}
	for c in  in_dict:
		res_tem[in_dict[c]['strain']]=in_dict[c]['cls_per']
		sc[in_dict[c]['strain']]=c
	res=sorted(res_tem.items(),key=lambda d:d[1],reverse=True)
	c=1
	for r in res:
		o.write(str(c)+'\t'+r[0]+'\t'+'C'+str(sc[r[0]])+'\t'+str(in_dict[sc[r[0]]]['cls_per'])+'\t'+str(in_dict[sc[r[0]]]['cls_ab'])+'\t'+str(in_dict[sc[r[0]]]['cls_cov'])+'\t'+str(in_dict[sc[r[0]]]['cls_covered_num'])+'/'+str(in_dict[sc[r[0]]]['cls_total_num'])+'\n')
		c+=1

	
def vote_strain_L2_batch(input_fq,db_dir,out_dir,ksize,res,l2,msn):
	check=check_L1_res(res)
	if check==1:
		print('- Only single cluster is identified, will not go to the 2nd layer identification ...')
		'''
		for r in res:
			pre='L1_'+'C'+str(r)
			os.system('cp '+out_dir+'/'+pre+'_StrainVote_cls.report '+out_dir+'/final_report.txt')
		'''
		generate_single_report(res,out_dir)
		exit()
	if len(res)==1:
		# No need to normalize y
		print('- Only 1 cluster is identified ...')
		for r in res:
			cls='C'+str(r)
			nd=db_dir+'/Kmer_Sets_L2/Kmer_Sets/C'+str(r)
			cls_out=out_dir+'/'+cls
			build_dir(cls_out)
			cls_ab=res[r]['cls_ab']
			cls_cov=res[r]['cls_cov']
			#overlap_kmr={}
			#union_kmr={}
			item=[input_fq,nd,cls_out,ksize,cls_ab,cls,cls_cov,list(res.keys()),l2,msn]
			#vote_strain_L2(input_fq,nd,cls_out,ksize,cls_ab,overlap_kmr,ok_percent)
			vote_strain_L2(item)
			os.system('cp '+cls_out+'/StrainVote.report '+out_dir+'/final_report.txt')

	else:
		# We need to normalize the y - Besides, parallel can be considered here
		print('- '+str(len(res))+' clusters are identified ...')
		# Get union k-mers and frequency ratio of each cluster
		#overlap_kmr,ok_percent=extract_kmr_ratio(db_dir,res)
		#final_res=defaultdict(lambda:{})
		all_in_list=[]
		#res_single={}
		for r in res:
			if not res[r]['strain']==0:continue
			#item=[]
			cls='C'+str(r)
			nd=db_dir+'/Kmer_Sets_L2/Kmer_Sets/C'+str(r)
			cls_out=out_dir+'/'+cls
			build_dir(cls_out)
			cls_ab=res[r]['cls_ab']
			cls_cov=res[r]['cls_cov']
			item=[input_fq,nd,cls_out,ksize,cls_ab,cls,cls_cov,list(res.keys()),l2,msn]
			all_in_list.append(item)
		print('- Parallel strain-level identification ...')
		for item in all_in_list:
			vote_strain_L2(item)
			#exit()
		'''
		
		pool=multiprocessing.Pool(processes=5)
		for item in all_in_list:
			result=pool.apply_async(vote_strain_L2,(item,))
		pool.close()
		pool.join()
		'''
	
		
		
		# Merge all output and generate final report
		print('- Generate final report ...')
		merge_res(out_dir,res)
def remove_1(dcount,res):
	py=[]
	for r in res:
		if r[0] not in dcount:
			py.append(0)
		else:
			if dcount[r[0]]==1:
				py.append(0)
			else:
				py.append(dcount[r[0]])
	return py
def trans(dcount,res):
	py=[]
	for r in res:
		if r[0] not in dcount:
			py.append(1)
		else:
			py.append(dcount[r[0]])
	return py


#def vote_strain_L2(input_fq,db_dir,out_dir,ksize,cls_ab,overlap_kmr,ok_percent):
def vote_strain_L2(item):
	input_fq=item[0]
	db_dir=item[1]
	out_dir=item[2]
	ksize=item[3]
	cls_ab=item[4]
	cls=item[5]
	cls_cov=item[6]
	all_cls=item[7]
	l2=item[8]
	msn=item[9]
	kid_match=pickle.load(open(db_dir+"/all_kid.pkl","rb"))
	#dk_match=pickle.load(open(db_dir+"/kmatch.pkl", "rb"))
	
	
	vote_d= defaultdict(lambda:0)# This dict is used to record the count score 
	vote_num=defaultdict(lambda:{})
	dir_jf=os.path.split(os.path.abspath(__file__))[0]+'/jellyfish-linux'
	# First Step --- Use Jellyfish to count kmers
	uid=uuid.uuid1().hex
	os.system(dir_jf+' count -m '+str(ksize)+' -s 100M -t 10 --if '+db_dir+'/all_kmer.fasta -o '+cls+'_'+uid+'.jf '+input_fq)	
	#print(dir_jf+' count -m '+str(ksize)+' -s 100M -t 8 --if '+db_dir+'/all_kmer.fasta -o Tem.jf '+input_fq)
	#exit()
	os.system(dir_jf+' dump -c '+cls+'_'+uid+'.jf > '+cls+'_'+uid+'.fa')
	#exit()
	# We will normalize y if there are multiple clusters identified! - Below.
	tcls=re.sub('C','',cls)
	tcls=int(tcls)
	'''
	if len(union_kmr)>1:
		dcount_o,total_kcount=load_kmer_count(cls+'_Tem.fa',overlap_kmr,ok_percent)	
	'''
	dcount_o,total_kcount=load_kmer_count(cls+'_'+uid+'.fa')
	os.system('rm '+cls+'_'+uid+'.jf '+cls+'_'+uid+'.fa')
	for k in kid_match:
		kid_match[k]=int(kid_match[k])
	res=sorted(kid_match.items(),key=lambda d:d[1])
	py_o=remove_1(dcount_o,res)
	#py_u=remove_1(dcount_u,res)
	#py_u=trans(dcount_u,res)
	'''
	for r in res:
		if r[0] not in dcount_o:
			py.append(0)
		else:
			if dcount_o[r[0]]==1:
				py.append(0)
			else:
				py.append(dcount[r[0]])
	'''
	#py_u=np.array(py_u)
	py_o=np.array(py_o)
	npp=py_o[py_o!=0]
	#npp25=1
	#npp25=np.percentile(npp,2,interpolation='nearest')
	npp25=0
	#npp25=np.min(npp)
	npp_outlier=np.median(npp)*1000
	#npp75=np.percentile(npp,98,interpolation='nearest')
	#print(npp_outlier,npp75,np.max(npp))
	#print(npp_outlier)
	#exit()
	npp75=npp_outlier
	#npp75=np.max(npp)
	#res,res2,strain_cov,strain_val,final_src=identify_strains_L2_Enet_Pscan_new.detect_strains(db_dir+'/all_strains_re.csv',py_o,db_dir+'/id2strain_re.pkl',int(ksize),npp25,npp75,npp_outlier,py_u,cls_cov)
	res,res2,strain_cov,strain_val,final_src=identify_strains_L2_Enet_Pscan_new_sp.detect_strains(db_dir+'/all_strains_re.npz',py_o,db_dir+'/id2strain_re.pkl',int(ksize),npp25,npp75,npp_outlier,cls_cov,db_dir+'/overlap_matrix.npz',all_cls,l2,msn)
	if len(res)==0:return
	
	nr=sorted(res.items(),key=lambda d:d[1],reverse=True)
	o=open(out_dir+'/StrainVote.report','w+')
	c=1
	o.write('Strain_ID\tStrain_Name\tCluster_ID\tRelative_Abundance_Inside_Cluster\tPredicted_Depth (Enet)\tPredicted_Depth (Ab*cls_depth)\tCoverage\tCoverd/Total_kmr\tValid_kmr\tRemain_Coverage\tCV\tExist_Evidence\n')
	# Recalculate the depth -> Don't consider those strains without exist evidence
	tdep=0
	for n in nr:
		#if n[1]>0.02 and strain_cov[n[0]][0]>0.7:
		tdep+=res2[n[0]]

	for n in  nr:
		if n[1]>0.02 and strain_cov[n[0]][0]>0.7:
			o.write(str(c)+'\t'+n[0]+'\t'+cls+'\t'+str(n[1])+'\t'+str(res2[n[0]])+'\t'+str((res2[n[0]]/tdep)*cls_ab)+'\t'+str(strain_cov[n[0]][0])+'\t'+str(strain_cov[n[0]][1])+'/'+str(strain_cov[n[0]][2])+'\t'+str(strain_val[n[0]])+'\t'+str(final_src[n[0]])+'\t*\n')
		else:
			o.write(str(c)+'\t'+n[0]+'\t'+cls+'\t'+str(n[1])+'\t'+str(res2[n[0]])+'\t'+str((res2[n[0]]/tdep)*cls_ab)+'\t'+str(strain_cov[n[0]][0])+'\t'+str(strain_cov[n[0]][1])+'/'+str(strain_cov[n[0]][2])+'\t'+str(strain_val[n[0]])+'\t'+str(final_src[n[0]])+'\t\n')
		c+=1
	o.close()

#vote_strain_L2_batch('../Real_Data_Record/Real_Data/SRR3184328.fq','../BacSkin_All','SRR3184328_Sep_Test',31,{3:{'cls_ab':41,'cls_cov':0.9998,'cls_covered_num':2286091,'cls_total_num':2286448,'cls_per':1,'s_ab':0,'strain':0}})

#vote_strain_L2_batch('../2021-03-31-Human_Skin_Real_Exp/S51.fq','../BacSkin_All','SRR3184328_Sep_Test',31,{2: {'cls_ab': 14.785894804912008, 'cls_cov': 0.06078654655387006, 'cls_covered_num': 303094, 'cls_total_num': 4986202, 'cls_per': 0.5797280522396767, 'strain': 'GCA_005768695', 's_ab': 14.785894804912008}, 1: {'cls_ab': 13.702900619522083, 'cls_cov': 0.01472852081953599, 'cls_covered_num': 70054, 'cls_total_num': 4756350, 'cls_per': 0.13399232242010173, 'strain': 'GCA_001062245', 's_ab': 13.702900619522083}, 14: {'cls_ab': 66.40591370588116, 'cls_cov': 0.49215618800858263, 'cls_covered_num': 48627, 'cls_total_num': 98804, 'cls_per': 0.09300888831932917, 'strain': 0, 's_ab': 0}, 6: {'cls_ab': 24.85808442855701, 'cls_cov': 0.362571865604726, 'cls_covered_num': 54992, 'cls_total_num': 151672, 'cls_per': 0.10518322714657598, 'strain': 0, 's_ab': 0}, 4: {'cls_ab': 173.68371087453005, 'cls_cov': 0.283064327772929, 'cls_covered_num': 46054, 'cls_total_num': 162698, 'cls_per': 0.02275857684839691, 'strain1': 'GCF_000302515', 's1_ab': 189.9782386884377, 's1_cov': 0.2840926522472772, 's1_covered_num': 18520, 's1_total_num': 65190, 's1_per': 0.0354232136811643, 'strain2': 'GCF_003611765', 's2_ab': 175.5555879292132, 's2_cov': 0.28238708618779995, 's2_covered_num': 27534, 's2_total_num': 97508, 's2_per': 0.052664296193152146}})



import re
import os
import numpy as np
from collections import defaultdict


def pick_rep(in_matrix,cls_file,out_dir):
	f1=open(in_matrix,'r')
	line=f1.readline().strip()
	ele=line.split('\t')
	#ida=range(len(ele))
	dc90=defaultdict(lambda:{})
	all_strains={}
	dpi={} # pre -> ID
	dpn={} # pre -> full dir name
	c=0
	for e in ele:
		pre=re.split('/',e)[-1]
		pre=re.split('\.',pre)[0]
		all_strains[pre]=''
		dpi[pre]=c
		dpn[pre]=e
		c+=1
	dpd={} # pre -> distance arr
	while True:
		line=f1.readline().strip()
		if not line:break
		ele=line.split('\t')
		pre=pre=re.split('/',ele[0])[-1]
		pre=re.split('\.',pre)[0]
		dpd[pre]=np.array(ele[1:])
	#print(sorted(dpd['GCF_001069395']))
	#exit()
	# Read cls file
	f2=open(cls_file,'r')
	o=open(out_dir+'/hclsMap_95_Rep.txt','w+')
	final_res={} # record all rep strains -> Rep -> Cid
	strain_rep={} # Strain -> Rep Strain
	clsa=[]
	while True:
		line=f2.readline().strip()
		if not line:break
		ele=line.split('\t')
		clsa.append(int(ele[0]))
		if int(ele[1])==1:
			final_res[ele[2]]=int(ele[0])
			strain_rep[ele[2]]=ele[2]
			dc90[int(ele[0])][dpn[ele[2]]]=''
			o.write(line+'\t0\n')
		elif int(ele[1])==2:
			rand=re.split(',',ele[-1])[0]
			final_res[rand]=int(ele[0])
			teme=re.split(',',ele[-1])
			dc90[int(ele[0])][dpn[rand]]=''
			for st in teme:
				strain_rep[st]=rand
			o.write(ele[0]+'\t'+ele[1]+'\t'+rand+'\t0\n')
		else:
			dres={} # Pre -> average distance
			dmix={} # Pre -> [Min, Max]
			sarr=re.split(',',ele[-1])
			#sarr2=re.split(',',ele[-1])
			for s in sarr:
				tem_arr=sarr.copy()
				tem_arr.remove(s)
				index=[]
				for t in tem_arr:
					index.append(dpi[t])
				index=np.array(index)
				dis=dpd[s][index]
				dis=dis.astype(np.float)
				dism=np.mean(dis)
				dres[s]=dism
				dmix[s]=[np.min(dis),np.max(dis)]
			res=sorted(dres.items(),key=lambda d:d[1])
			win=res[0][0]
			dc90[int(ele[0])][dpn[win]]=''
			final_res[win]=int(ele[0])
			for s in sarr:
				strain_rep[s]=win
			o.write(ele[0]+'\t'+ele[1]+'\t'+res[0][0]+'\t'+str(dmix[win][0])+','+str(dmix[win][1])+','+str(dres[win])+'\n')
	o2=open(out_dir+'/Other_Strain_CN.txt','w+')
	index={}
	cls_ns=defaultdict(lambda:{}) # Cls -> new strains in this cluster
	for s in final_res:
		index[s]=dpi[s]
	#index=np.array(index)
	for a in all_strains:
		if a in final_res:
			cls_ns[final_res[a]][a]=''
			continue
		# Distance to all rep strains
		dtem={}
		for s in index:
			dtem[s]=float(dpd[a][index[s]])
		#dis=dpd[a][index]
		#dis=dis.astype(np.float)
		close=sorted(dtem.items(),key=lambda d:d[1])
		if close[0][0]==strain_rep[a]:
			cls_ns[final_res[strain_rep[a]]][a]=''
			continue
		cls_ns[final_res[close[0][0]]][a]=''
		o2.write(a+'\t'+strain_rep[a]+','+str(dtem[strain_rep[a]])+'\t'+close[0][0]+','+str(dtem[close[0][0]])+'\n')
		'''
		if not strain_rep[a]==close[0][0]:
			print(a)
		'''
	dc90_l2=defaultdict(lambda:{})
	o3=open(out_dir+'/hclsMap_95_recls.txt','w+')
	for i  in clsa:
		o3.write(str(i)+'\t'+str(len(cls_ns[i]))+'\t'+','.join(list(cls_ns[i].keys()))+'\n')
		for strain in list(cls_ns[i].keys()):
			dc90_l2[int(i)][dpn[strain]]=''
	return dict(dc90),dict(dc90_l2)
	
	
	

#a=pick_rep('../../StrainVote/BacSkin_All/Cluster_Result/distance_matrix_rebuild.txt','../../StrainVote/BacSkin_All/Cluster_Result/hclsMap_single_95.txt','../../StrainVote/BacSkin_All/Cluster_Result')



class tree_node():
    def __init__(self):
        self.seq = -1
        self.child = []
        self.father = None
        self.category = 2 # 0, 'o1', 'o2', 1, 2: 2 -> (1000, )
        self.leaves = []
        self.access = 0
        self.cov_correct = 0
        self.res_proc_label = 0
        # output format
        self.total_num = 0
        self.covered_num = 0
        self.strain = ""
    def get_child(self, T1, T2):
        self.child.append(T1)
        self.child.append(T2)
        T1.father = self
        T2.father = self

import scipy.stats as st
import time
import os
import random
import uuid
import numpy as np
from collections import defaultdict

def seq_node(seq, nodes):
    for i in nodes:
        if(i.seq == seq):
            return i

def read_tree(nodes, leaves, db_dir):
    f = open(db_dir+"/tree_structure.txt", "r")
    lines = f.readlines()
    for line in lines:
        temp = line.rstrip().split("\t")
        T = tree_node()
        T.seq = int(temp[0])
        if(temp[1] == "N"):
            T0 = T
        else:
            T.father = int(temp[1])
        nodes.append(T)
        if(temp[2] != "N"):
            temp1 = temp[2].split(" ")
            T.child = [int(temp1[0]), int(temp1[1])]
        else:
            leaves.append(T)
        temp1 = temp[3].split(" ")
        if('' in temp1):
            del temp1[-1]
        #print(temp1)
        temp1[-1] = temp1[-1].rstrip()
        for i in temp1:
            T.leaves.append(int(i))
        if(len(temp) == 5):
            T.strain = temp[4]
    # get true c&f
    for i in nodes:
        if(i.father != None):
            i.father = seq_node(i.father, nodes)
        if(i.child != []):
            temp = []
            for j in i.child:
                temp.append(seq_node(j, nodes))
            i.child = temp
    return T0

def nodes_classify(nodes, db_dir, length_cutoff):
    length = {}
    f = open(db_dir+"/nodes_length.txt", "r")
    lines = f.readlines()
    for line in lines:
        d = line.rstrip().split("\t")
        length[int(d[0])] = int(d[1])
    for i in nodes:
        if(length[i.seq] < length_cutoff):
            i.category = 0
        elif(length[i.seq] < 1000):
            i.category = 1
    f = open(db_dir + "/overlap_nodes", "r")
    lines = f.readlines()
    for line in lines:
        x = seq_node(int(line.rstrip()), nodes)
        if(x.category != 0):
            if(length[x.seq] < 10000):
                x.category = 'o1'
            else:
                x.category = 'o2'

def jellyfish_count(fq_path, db_dir, match_results):
    if(type(fq_path) == str):
        x = fq_path
    else:
        x = " ".join(fq_path)
    dir_jf = os.path.split(os.path.abspath(__file__))[0]+'/jellyfish-linux'
    #path = "temp.fa"
    #if(os.path.exists(path) == False):
    #    os.system(dir_jf + " count -m 31 -s 100M -t 8 --if " + db_dir+"/kmer.fa -o temp.jf " + x)
    #    os.system(dir_jf+" dump -c temp.jf > temp.fa")
    #    os.system("rm temp.jf")
    uid = uuid.uuid1().hex
    path = "temp_"+uid+".fa"
    os.system(dir_jf + " count -m 31 -s 100M -t 10 --if " + db_dir+"/kmer.fa -o temp_"+uid+".jf " + x)
    os.system(dir_jf+" dump -c temp_"+uid+".jf > temp_"+uid+".fa")
    os.system("rm temp_"+uid+".jf")
    f = open(path, "r")
    lines = f.readlines()
    for line in lines:
        temp = line.rstrip().split(" ")
        match_results[temp[0]] = int(temp[1])
    os.system("rm temp_"+uid+".fa")

def del_outlier(profile):
    delete_index = []
    x = 100 * np.median(profile)
    for i in range(0, len(profile)):
        if(profile[i]>=x):
            delete_index.append(i)
    delete_index.sort(reverse = True)
    for i in delete_index:
        del profile[i]

def match_node(node, profile, match_results, db_dir):
    f = open(db_dir+"/nodes_kmer/"+str(node.seq), "r")
    lines = f.readlines()
    uncovered = 0
    line = lines[0].rstrip().split(" ")
    for x in line:
        if(x in match_results):
            if(match_results[x]>0): # dont ignore 1
                profile.append(match_results[x])
            else:
                uncovered += 1
    length = uncovered+len(profile)
    if(len(profile)>0):
        del_outlier(profile)
    return length

def piecewise(cov_cutoff, cov, label, profile):
    if(cov == 0):
        return 0
    if(label in [1, 'o1']):
        x = cov_cutoff[1]
    elif(label in [2, 'o2']):
        x = cov_cutoff[0]
    if(cov >= x):
        return np.mean(profile)
    else:
        return 0

def get_unique_kmer(node, nodes, cov, abundance, length, results, length_cutoff, cov_cutoff, db_dir, match_results):
    delete = set([])
    profile = []
    overlap = defaultdict(list)
    uncovered = 0
    f = open(db_dir+"/nodes_kmer/"+str(node.seq), "r")
    lines = f.readlines()
    line = lines[0].rstrip().split(" ")
    for i in results:
        path = db_dir+"/overlap/"+str(node.seq)+"_"+str(i.seq)
        if(os.path.exists(path) == False):
            continue
        f = open(path, "r")
        lines = f.readlines()
        if(len(lines)>0):
            overlap[i.seq] = list(map(int, lines[0].rstrip().split(" ")))
            delete = set(overlap[i.seq]) | delete
    if(len(line) - len(delete) >= length_cutoff):
        remain = set(list(range(0, len(line)))) - delete
        remain = list(remain)
        for k in remain:
            if(line[k] in match_results):
                if(match_results[line[k]]>0):
                    profile.append(match_results[line[k]])
                else:
                    uncovered += 1
        length[node.seq] = uncovered + len(profile)
        if(length[node.seq]<length_cutoff):
            abundance[node.seq] = 0
            cov[node.seq] = 1
            length[node.seq] = 0
            node.category = 0
            return 0
        if(len(profile)>0):
            del_outlier(profile)
        cov[node.seq] = len(profile)/length[node.seq]
        if(length[node.seq]<1000):
            x = 1
            abundance[node.seq] = piecewise(cov_cutoff, cov[node.seq], 1, profile)
        else:
            x = 2
            abundance[node.seq] = piecewise(cov_cutoff, cov[node.seq], 2, profile)
        return x
    else:
        temp_match = {}
        j = 0
        node.cov_correct = 1
        for x in line:
            if(x in match_results):
                temp_match[j] = match_results[x]
            j += 1
        x = {}
        temp = []
        for j in results:
            #x[j.seq] = abundance[j.seq]
            x[j.seq] = len(overlap[j.seq])
        Tuple = sorted(x.items(), key = lambda kv:(kv[1], kv[0]), reverse = True)
        for k in Tuple:
            temp.append(seq_node(k[0], nodes))
        results = temp.copy()
        for j in results:
            temp_match1 = {}
            if(j.seq in overlap):
                for k in overlap[j.seq]:
                    x = int(k)
                    if(temp_match[x]>0):
                        temp_match1[x] = temp_match[x]
            sample = np.random.poisson(abundance[j.seq], size=len(temp_match1))
            sample.sort()
            Tuple = sorted(temp_match1.items(), key = lambda kv:(kv[1], kv[0]))
            #de = 1 - cov[j.seq]
            #DE = list(np.linspace(0, len(sample), num = int(de*len(sample))))
            #DE = set(map(int, DE))
            #rem = list(set(range(0, len(sample))) - DE)
            for k in range(0, len(sample)):
            #for k in rem:
                temp_match[Tuple[k][0]] = Tuple[k][1] - sample[k]
        for j in temp_match:
            if(temp_match[j]>0):
                profile.append(temp_match[j])
        length[node.seq] = len(temp_match)
        del_outlier(profile)
        cov[node.seq] = len(profile)/length[node.seq]
        if(len(line)<1000):
            x = 'o1'
            abundance[node.seq] = piecewise(cov_cutoff, cov[node.seq], x, profile)
        else:
            x = 'o2'
            abundance[node.seq] = piecewise(cov_cutoff, cov[node.seq], x, profile)
        return x

def process_zero_ab(T, results, abundance, length, cov, ab_cutoff, zero_label):
    x = T.father
    while(x.father!=None):
        y = get_father_ab(x, length, cov, abundance)
        if(y >= 1):
            break
        x = x.father
    ab = y
    res = []
    zero_num = 0
    for i in results:
        if(i.seq in x.leaves):
            if(zero_label[i.seq] == 1):
                zero_num += 1
            res.append(i.seq)
    if(zero_num == 1):
        for i in res:
            if(i != T.seq):
                ab = ab - abundance[i]
    else:
        return abundance[T.seq]
    if(ab >= ab_cutoff):
        return ab
    else:
        return 0

def get_uniq_path(T, path):
    path.append(T)
    if(T.father == None or find_bro(T).access in [1, 2]):
        return 1
    get_uniq_path(T.father, path)

def get_father_ab(T, length, cov, abundance):
    path = []
    get_uniq_path(T, path)
    Sum = 0
    l = 0
    for j in path:
        Sum += length[j.seq] * cov[j.seq]
        l += length[j.seq]
    if(l != 0):
        ratio = []
        ab = []
        for j in path:
            ratio.append(cov[j.seq]*length[j.seq]/Sum)
            ab.append(abundance[j.seq])
        return sum([a*b for a,b in zip(ab,ratio)])
    else:
        if(path[-1].father == None): # root node empty
            return -1
        if(find_bro(path[-1]).access == 2):
            return abundance[path[-1].father.seq]
        else:
            return abundance[path[-1].father.seq] - abundance[find_bro(path[-1]).seq]

def check_access(T, label):
    T.access = label
    if(label == 1 and T.father == None):
        return 1
    elif(label == 0 and (T.father == None or find_bro(T).access in [1, 2])):
        return 1
    check_access(T.father, label)

def identification(nodes, pending, db_dir, match_results, cov_cutoff, length, abundance, cov, length_cutoff, leaves, results, ab_cutoff):
    group = pending[0]
    print("--------------------------------------------------")
    if(len(group) == 1 and group[0].category in [1, 2, 'o1', 'o2']): # assume T0 is 2 or 1
        profile = []
        x = group[0]
        group[0].access = 1
        length[x.seq] = match_node(x, profile, match_results, db_dir)
        if(length[x.seq]!=0):
            cov[x.seq] = len(profile)/length[x.seq]
        else:
            cov[x.seq] = 0
        abundance[x.seq] = piecewise(cov_cutoff, cov[x.seq], x.category, profile)
        print(x.seq)
        print(abundance[x.seq], end = " | ")
        print(cov[x.seq])
        if(abundance[x.seq] > 1):
            pending.append((x.child[0], x.child[1]))
        pending.remove(group)
        return 1
    elif(len(group) == 1 and group[0].category == 0): # do not use it
        x = group[0]
        x.access = 1
        length[x.seq] = 0
        cov[x.seq] = 0
        abundance[x.seq] = 0
        print(x.seq)
        print(abundance[x.seq], end = " | ")
        print(cov[x.seq])
        pending.append((x.child[0], x.child[1]))
        pending.remove(group)
        return 1
    if(group[0].category == 0 and group[1].category == 0):
        print("%d\ndefault\n%d\ndefault"%(group[0].seq, group[1].seq))
        group[0].access = 2
        group[1].access = 2
        for i in group:
            if(i in leaves):
                results.append(i)
            else:
                pending.append((i.child[0], i.child[1]))
            length[i.seq] = 0
            cov[i.seq] = 0
            # test
            abundance[i.seq] = abundance[i.father.seq]
        pending.remove(group)
        return 1
    X = []
    for i in group:
        if(i.category == 0):
            abundance[i.seq] = 0
            cov[i.seq] = 0.85
            length[i.seq] = 0
            X.append((i, 0))
            print("%d\ndefault"%i.seq)
        elif(i.category in [1, 2] or len(results) == 0):
            if(i.category == 'o1'):
                i.category = 1
            elif(i.category == 'o2'):
                i.category = 2
            X.append((i, i.category))
            profile = []
            length[i.seq] = match_node(i, profile, match_results, db_dir)
            if(length[i.seq]>0):
                cov[i.seq] = len(profile)/length[i.seq]
            else:
                cov[i.seq] = 0
            abundance[i.seq] = piecewise(cov_cutoff, cov[i.seq], i.category, profile)
            if(abundance[i.seq] < ab_cutoff):
                abundance[i.seq] = 0
            print("%d\n%f | %f"%(i.seq, abundance[i.seq], cov[i.seq]))
        else:
            label = get_unique_kmer(i, nodes, cov, abundance, length, results, length_cutoff, cov_cutoff, db_dir, match_results)
            if(label != 0):
                print("%d\n%f | %f"%(i.seq, abundance[i.seq], cov[i.seq]))
            if(abundance[i.seq] < ab_cutoff):
                abundance[i.seq] = 0
            X.append((i, label))
    if(X[0][1] == 0 and X[1][1] == 0):
        return 1
    # correctness, x need correct
    if(X[0][1] in [1, 2] and X[1][1] in [1, 2]):
        pass
    elif(set([X[0][1], X[1][1]]) == [set(['o2', 1])]):
        pass
    else:
        label = 0
        abundance_t = get_father_ab(group[0].father, length, cov, abundance)
        if(set([X[0][1], X[1][1]]) in [set(['o1', 2])]):
            for i in X:
                if(i[1] == 2):
                    y = i[0]
                else:
                    x = i[0]
        elif(0 in set([X[0][1], X[1][1]])):
            for i in X:
                if(i[1] == 0):
                    x = i[0]
                else:
                    y = i[0]
            label = 2
        elif(set([X[0][1], X[1][1]]) == set(['o1', 'o2'])):
            for i in X:
                if(i[1] == 'o1'):
                    x = i[0]
                else:
                    y = i[0]
        elif(set([X[0][1], X[1][1]]) in [set(['o1', 'o1']), set(['o2', 'o2'])]):
            label = 1
        else:
            label = - 1
        if(abundance_t == -1): # root empty
            access = [x]
            if(abundance[y.seq] >= ab_cutoff):
                access.append(y)
                y.access = 1
            for i in access:
                if(i == x):
                    length[i.seq] = 0
                    abundance[i.seq] = 0
                    cov[i.seq] = 0
                    i.access = 2
                if(i in leaves):
                    results.append(i)
                else:
                    pending.append((i.child[0], i.child[1]))
            pending.remove(group)
            return 1
        if(label == 0):
            abundance[x.seq] = abundance_t - abundance[y.seq]
            abundance[x.seq] = piecewise(cov_cutoff, cov[x.seq], x.category, [abundance[x.seq]])
        elif(label == 1):
            for i in [X[0][0], X[1][0]]:
                abundance[i.seq] = abundance_t * (abundance[i.seq]/(abundance[i.seq]+abundance[find_bro(i).seq]))
        elif(label == 2): # 0
            abundance[x.seq] = abundance_t - abundance[y.seq]
    # binomial test
    ab_temp = {}
    for i in range(0, 2):
        if(abundance[group[i].seq]<=0):
            ab_temp[group[i]] = 0
        else:
            ab_temp[group[i]] = round(abundance[group[i].seq])
    if(list(ab_temp.values()) == [0, 0]):
        pending.remove(group)
        return 1
    Tuple = sorted(ab_temp.items(), key = lambda kv:(kv[1]))
    (a, b, x, y) = (Tuple[1][0], Tuple[0][0], Tuple[1][1], Tuple[0][1])
    ret = 1 - st.binom.sf(max([x, y]), x+y, 0.995)
    if(ret < 0.05):
        for i in (a, b):
            i.access = 1
            if(i.category!=0):
                check_access(i, 1)
            if(i not in leaves):
                pending.append((i.child[0], i.child[1]))
            else:
                results.append(i)
    else:
        if(a.category!=0):
            check_access(a, 1)
        a.access = 1
        if(a not in leaves):
            pending.append((a.child[0], a.child[1]))
        else:
            results.append(a)
        check_access(b, 0)
    pending.remove(group)
    return 1

def find_bro(T):
    for i in T.father.child:
        if(i!=T):
            return i

def res_node_proc(node, abundance, length, cov, overall_cutoff, zero_label):
    node.res_proc_label = 1
    path_t = []
    path = []
    get_uniq_path(node, path_t)
    for j in path_t:
        if(j.cov_correct == 0 and length[j.seq]!=0):
            path.append(j)
    if(len(path) == 0):
        path = path_t.copy()
    zero_ab = []
    for j in path:
        zero_ab.append(abundance[j.seq])
        node.covered_num += length[j.seq]*cov[j.seq]
        node.total_num += length[j.seq]
    node.covered_num = int(node.covered_num)
    if(node.total_num != 0 and node.covered_num/node.total_num < overall_cutoff):
        return 0
    ratio = []
    ab = []
    if(node.total_num != 0):
        for j in path:
            ratio.append(cov[j.seq]*length[j.seq]/node.covered_num)
            ab.append(abundance[j.seq])
        abundance[node.seq] = sum([a*b for a,b in zip(ab,ratio)])
        zero_label[node.seq] = 0
    else:
        zero_label[node.seq] = 1
    return 1

def get_results(results, cov, length, abundance, overall_cutoff, ab_cutoff, zero_label):
    total_ab = 0
    delete = []
    for i in results:
        if(zero_label[i.seq] == 1):
            ab_t = process_zero_ab(i, results, abundance, length, cov, ab_cutoff, zero_label)
            if(ab_t == 0):
                delete.append(i)
                abundance[i.seq] == 0
            else:
                abundance[i.seq] = ab_t
        total_ab += abundance[i.seq]
    for i in delete:
        results.remove(i)
    return total_ab

def identify_cluster(fq_path, db_dir, cutoff):
    # parameters
    #cov_cutoff = [0.12, 0.1] # >1000, (350, 1000)
    cov_cutoff = [cutoff[0], cutoff[0]]
    length_cutoff = 350 # minimum k-mers of nodes, < 1000

    start=time.time()

    # initialization
    nodes = []
    leaves = []
    match_results = {}
    T0 = read_tree(nodes, leaves, db_dir)
    nodes_classify(nodes, db_dir, length_cutoff)
    jellyfish_count(fq_path, db_dir, match_results)

    pending = [[T0]]
    results = []
    length = {}
    abundance = {}
    cov = {}
    zero_label = {}
    delete = []
    while(len(pending) != 0):
        identification(nodes, pending, db_dir, match_results, cov_cutoff, length, abundance, cov, length_cutoff, leaves, results, cutoff[2])
        for j in results:
            if(j.res_proc_label == 0):
                x = res_node_proc(j, abundance, length, cov, cutoff[1], zero_label)
                if(x == 0):
                    delete.append(j)
    results_copy = results.copy()
    for j in delete:
        results.remove(j)
    if(len(results) > 0):
        total_ab = get_results(results, cov, length, abundance, cutoff[1], cutoff[2], zero_label)
    elif(len(results_copy) != 0):
        cov_list = {}
        for j in results_copy:
            cov_list[j] = j.covered_num/j.total_num
        r = max(cov_list, key=cov_list.get)
        if(cov_list[r]>=0.1):
            results = [r]
            total_ab = abundance[r.seq]
        else:
            results = []
    # output
    res = defaultdict(lambda:{})
    for i in results:
        res[i.seq]['cls_ab'] = abundance[i.seq]
        res[i.seq]['cls_per'] = abundance[i.seq]/total_ab
        if(i.total_num == 0):
            res[i.seq]['cls_cov'] = -1
        else:
            res[i.seq]['cls_cov'] = i.covered_num/i.total_num
        res[i.seq]['cls_total_num'] = i.total_num
        res[i.seq]['cls_covered_num'] = i.covered_num
        res[i.seq]['strain'] = 0
        res[i.seq]['s_ab'] = 0
        if(i.strain != ""):
            res[i.seq]['strain'] = i.strain
            res[i.seq]['s_ab'] = abundance[i.seq]
    end = time.time()
    print('- The total running time of identification is ',str(end-start),' s\n')
    return res
'''
for i in range(7, 50):
    os.system("rm temp.fa")
    os.system("python simulate.py "+ str(i))
    fq = ("reads/1.fq", "reads/2.fq")
#fq = ("/home/heruiliao2/Bacteria_Genome_Graph/StrainVote_Rep_Uniq_Exp/Sim_Short_Data/bac8.fq")
#results = identify_cluster(fq, "/home/yongxinji2/worktemp/Tree_database")
    results = identify_cluster(fq, "Tree_database")
    for i in results.keys():
        print(i, results[i])

#fq = ("reads/1.fq", "reads/2.fq")
fq = ("/home/heruiliao2/Bacteria_Genome_Graph/Benchmark_Tools/All_Real_Data/Pre_Real/SRR8146961.fq")
#fq = ("/home/heruiliao2/Bacteria_Genome_Graph/StrainVote_Rep_Uniq_Exp/Sim_Short_Data/bac8.fq")

#fq = ("/home/heruiliao2/Bacteria_Genome_Graph/Benchmark_Tools/QuantTB/db_Mtb/Real_Data/ERR171163.fastq")
#fq = ("/home/heruiliao2/Bacteria_Genome_Graph/Benchmark_Tools/All_Real_Data/Cae_Real/MET0191.fq")
#fq = ("/home/heruiliao2/Bacteria_Genome_Graph/Benchmark_Tools/All_Real_Data/Cae_Real/MET0404.fq")
#fq = ("/home/heruiliao2/Bacteria_Genome_Graph/Benchmark_Tools/All_Real_Data/Cae_Real/MET0423.fq")
#fq = ("/home/heruiliao2/Bacteria_Genome_Graph/Benchmark_Tools/All_Real_Data/Ecoil_Real/S2840_39x.fastq")
#fq = ("/home/heruiliao2/Bacteria_Genome_Graph/Benchmark_Tools/All_Real_Data/Ecoil_Real/S3751_10x.fastq")
#fq = "/home/heruiliao2/Bacteria_Genome_Graph/Benchmark_Tools/All_Real_Data/HMP_mock/SRR172902.fastq"

#results = identify_cluster(fq, "Tree_database_ecoli")
#cutoff = [0.05, 0.1, 1]
cutoff = [0.12, 0.5, 2]
#cutoff = (0.05, 0.6, 1)
#cutoff = [0.05, 0.1, 1]
sp = "Pre"
results = identify_cluster(fq, "Lib/"+sp, cutoff)
for i in results.keys():
    print(i, results[i])

'''



import re
import os
from collections import defaultdict
import seqpy
from Bio import SeqIO
#from Bio import pairwise2


def get_pre(filename):
	name=os.path.split(filename)[1]
	pre=re.split('\.',name)[0]
	return pre


def parse_sts_res(sts_res,ksize,dlabel,sid_match,kid_match,knum,kmatrix,label_match):
	# Get the number of strains
	ifile=(sts_res)
	ifile2=('reference.fasta')
	num=0
	dmsa={} # Seq_Name -> msa
	my_dict = SeqIO.to_dict(SeqIO.parse(ifile, "fasta"))
	my_dict2=SeqIO.to_dict(SeqIO.parse(ifile2, "fasta"))

	for r in my_dict:
		num+=1
		seqn=re.sub('\..*','',r)
		seq=str(my_dict[r].seq)
		dmsa[seqn]=seq
	for r in my_dict2:
		num+=1
		seqn=re.sub('\..*','',r)
		seq=str(my_dict2[r].seq)
		dmsa[seqn]=seq

	'''
	o=open('msa_res.aln','w+')
	while True:
		line=f.readline().strip()
		if not line:break
		
		if line[0]=='s':
			ele=line.split()
			name=re.split('\.',ele[1])[-1]
			seq=ele[-1]
			dmsa[name]=seq
			o.write('>'+name+'\n'+seq+'\n')
		
		if re.search('>',line):
			name=re.sub('>','',line)
			dmsa[name]=''
			#num+=1
		else:
			dmsa[name]+=line
	'''
		
	fdir=os.path.split(os.path.realpath(__file__))[0]
	pwd=os.getcwd()
	#print('perl '+fdir+'/aln2entropy.pl '+pwd+'/msa_res.aln '+str(len(dmsa))+' > tem_column.txt')
	os.system('strainest map2snp reference.fasta '+sts_res+' snp.dgrp')
	#os.system('perl '+fdir+'/aln2entropy.pl '+pwd+'/msa_res.aln '+str(len(dmsa))+' > tem_column.txt')
	#exit()
	# Parse each column
	dash_cutoff=int(0.1*num) # 100 -> <=10 dash is ok
	#print(dash_cutoff,num)
	f2=open('snp.dgrp','r')
	line=f2.readline()
	out_c=[]
	while True:
		line=f2.readline().strip()
		if not line:break
		ele=re.split(',',line)
		cid=ele[0]
		#dash_num=int(re.sub('-:','',ele[2]))
		btype={}
		dash_num=0
		for e in ele[2:]:
			if 'N'==e:
				dash_num+=1
			else:
				btype[e]=''
				#bn=re.sub('.*:','',e)
				#bn=int(bn)
				#if bn>0:btype+=1
		if dash_num>dash_cutoff:continue
		#print(btype)
		if len(btype)>1:
			out_c.append(int(cid))
	#print(out_c)
	#exit()
	print('Log: Extract kmers from msa..., k-mer count:',len(kid_match),len(kmatrix))
	#for s in dmsa:
	#tc=0
	for c in out_c:
		tem_kmr={}
		go=1
		for s in dmsa:
			#print(s,c,dmsa[s][c],dmsa[s][c-1])
			if dmsa[s][c-1]=='N':
				#go=0
				continue
			kmr=extract_kmers(c-1,dmsa[s],ksize)
			if re.search('N',kmr):
				continue
				#go=0
			if not len(kmr)==ksize:go=0
			if len(dlabel[kmr])==0:go=0
			rev_kmr=seqpy.revcomp(kmr)
			tem_kmr[rev_kmr]=''
			tem_kmr[kmr]=''
		if go==0:continue
		if len(tem_kmr)==0:continue
		for kmr in tem_kmr:
			if kmr in kid_match:continue
			if kmr not in dlabel:continue
			#if len(dlabel[kmr])==1:continue
			if len(dlabel[kmr])>=len(sid_match):continue
			if len(dlabel[kmr])==0:continue
			kid_match[kmr]=knum
			for e in dlabel[kmr]:
				kmatrix[knum][sid_match[label_match[e]]-1]=1
			#print(knum,len(kmatrix),len(kid_match))
			knum+=1
			'''
			knum+=1
			kid_match[rev_kmr]=knum
			for e in dlabel[rev_kmr]:
				kmatrix[knum][sid_match[label_match[e]]-1]=1
			knum+=1
			'''
	print('Log: Extraction done..., k-mer count:',len(kid_match),len(kmatrix))
	os.system('rm -rf '+os.getcwd()+'/ref '+os.getcwd()+'/output')
	#exit()
	
		
		
def parse_seq(input_seq,sseq):
	f=open(input_seq,'r')
	s=''
	while True:
		line=f.readline().strip()
		if not line:break
		if re.search('>',line):
			s=re.sub('>','',line)
			sseq[s]=''
		else:
			sseq[s]=line
	return s



def extract_kmers(index,in_seq,ksize):
	#exit()
	if index>len(in_seq):
		print('Check!')
		exit()
	base=in_seq[index]
	left=in_seq[:index][::-1]
	right=in_seq[index+1:]
	lseq=''
	rseq=''
	for l in left:
		if len(lseq)==(ksize-1)/2:
			break
		if l=='-':continue
		if l=='N':break
		lseq+=l
	for r in right:
		if len(rseq)==(ksize-1)/2:
			break
		if r=='-':continue
		if r=='N':break
		rseq+=r
	lseq=lseq[::-1]
	kmr=lseq+base+rseq
	if len(kmr)<ksize:
		if len(lseq)<(ksize-1)/2:
			rseq=''
			rl=ksize-1-len(lseq)
			for r in right:
				if len(rseq)==rl:
					break
				if r=='-':continue
				if r=='N':break
				rseq+=r
		if len(rseq)<(ksize-1)/2:
			lseq=''
			ll=ksize-1-len(rseq)
			for l in left:
				if len(lseq)==ll:
					break
				if l=='-':continue
				if l=='N':break
				lseq+=l
		kmr=lseq+base+rseq
	'''
	if not len(kmr)==ksize:
		print(len(kmr),index,kmr,lseq,base,rseq)
		print('Kmer extraction --- something wrong! Check!')
		exit()
	'''
	return kmr
	
	

def align_with_sts(strains,msa_out,ksize,dlabel,sid_match,kid_match,knum,kmatrix,label_match):
	# Input -> One block from sub-maf
	# Output -> two dict -> drq and ikmer
	#black_list={} # Ref: {Ref_Index_in_genome : ''} -> These columns contain dash
	#base_type=defaultdict(lambda:{}) # Ref: {Ref_Index1_in_genome: {A:'',T:''}, Ref_Index2_in_genome:{T:''}}-> Used to filter columns that are completely similar
	#kmeri=defaultdict(lambda:{}) # Strain -> {seq_index_in_genome:index_in_block_seq} # We need this dict to extract kmers
	#sseq={} # Strain -> seq_seq # Used to extract kmers
	#tem_drq=defaultdict(lambda:{}) # Same structure to drq, but index means block seq index
	#tem_drq=defaultdict(lambda:{})
	#refname=parse_seq(ref,sseq) # Load seq of ref
	#i=0
	#tcount=1
	ast=' '.join(strains)
	fdir=os.path.split(os.path.realpath(__file__))[0]
	print('Log: Running Sts.Mapgenomes to align genomes...')
	#os.system('mafft --quiet --6merpair --thread -1  --addfragments '+query_name+' '+ref+' > tem_mafft.aln')
	#os.system('mafft --thread -1 --auto '+query_name+' > tem_mafft.aln')
	#os.system(fdir+'/muscle -align '+query_name+' -output tem_mafft.aln')
	os.system('strainest mapgenomes '+ast+' reference.fasta '+msa_out+'/mapped.fna')
	os.system('strainest map2snp reference.fasta '+msa_out+'/mapped.fna snp.dgrp')
	#os.system('mugsy --directory '+msa_out+' '+ast)
	print('Log: Alignment done...')	
	parse_sts_res(msa_out+'/mapped.fna',ksize,dlabel,sid_match,kid_match,knum,kmatrix,label_match)


	# When we finish this part, we can extract kmers
	'''
	for index in list(tem_drq):
		if len(base_type[index])==1: # Remove those columns with same bases
			drq.pop(index, None)		
			continue
		 	# Skip column
		# Continue to extract kmers
		rindex=int(index)-int(ref_start)
		#print(rindex,index,ref_start)
		#ikmer[refname][index]=extract_kmers(rindex,sseq[refname],ksize)
		inbase=str(index)+'-'+drq[index][refname][1]
		ikmer[extract_kmers(rindex,sseq[refname],ksize)][inbase]=''
		for qe in drq[index]:
			if qe not in query_start:continue
			qindex=int(drq[index][qe][0])-int(query_start[qe])
			#ikmer[qe][drq[index][qe][0]]=extract_kmers(qindex,sseq[qe],ksize)
			inbase=str(index)+'-'+drq[index][qe][1]
			ikmer[extract_kmers(qindex,sseq[qe],ksize)][inbase]=''
			rk=seqpy.revcomp(extract_kmers(qindex,sseq[qe],ksize))
			ikmer[rk][inbase]=''
	'''
			
	
	

		


def extract_kmer_sts(in_block,ksize,kid_match,kmatrix,knum,sid_match,dlabel,label_match):
	input_block=in_block
	#block_pre=get_pre(input_block) # For example, B1, B2 and B3
	#block_pre='Cls_'+block_pre
	#out_dir=input_list[1] # OutPut_dir -> Kmer_Sets dir
	#mcgr=input_list[2] # The strain with maximum completeness
	ksize=int(ksize) # Kmer size
	f=open(input_block,'r')
	lines=f.read().split('\n')
	c=0
	#o=open(out_block,'w+')
	#ikmer=defaultdict(lambda:{}) # Strain -> {index -> kmer} # Index here refers to index in the genome
	#dstrain={}
	showp=0#
	# We will connect all blocks with N and align them using mafft
	all_seq={} # { strain1 -> [seq1,seq2], strain2 ->,...}
	dlength={}
	print('Log: Connect all blocks...')
	#query_seq={} # Query -> [seq1, seq2]
	for line in lines:
		if not line:continue
		'''
		if not line:
			
			if not c==0:
				o.write('\n')
			
			continue
		'''
		if line[0]=='a':
			showp+=1
			#print('------Process:',showp)
			if not c==0 and not len(fa)==1:
				temc=1
				query_list=[]
				query_start={} # Strain name -> start-pos
				ref_start=0
				ref=''
				res=sorted(dltem.items(), key = lambda kv:(kv[1], kv[0]), reverse = True)
				if not res[0][1]<50: # Length filter
					#o2=open('other_ref.fasta','w+')
					#query_name='other_ref.fasta'
					for strain in fa:
						if strain not in all_seq:
							all_seq[strain]=[fa[strain]]
						else:
							all_seq[strain].append(fa[strain])
						'''
						if strain==res[0][0]:
							#ref='tem_Ref.fasta'
							if strain not in all_seq:
								all_seq[strain]=[fa[strain]]
							else:
								ref_seq[strain].append(fa[strain])
							#o1=open('tem_Ref.fasta','w+')
							#o1.write('>'+strain+'\n'+fa[strain]+'\n')
							#o1.close()
							ref_start=int(block_info[strain].split()[2])
						else:
							#query=block_pre+'_tem_'+str(temc)+'.fasta'
							#query_list.append(query)
							if strain not in query_seq:
								query_seq[strain]=[fa[strain]]
							else:
								query_seq[strain].append(fa[strain])
							query_start[block_info[strain].split()[1]]=int(block_info[strain].split()[2])
							#o2=open(query,'w+')
							#o2.write('>'+strain+'\n'+fa[strain]+'\n')
							#o2.close()
							#temc+=1
						'''
					#print('Ref: ',ref)
					#o1.close()
					#o2.close()
					#align_with_mafft(ref,query_name,ksize,dlabel,sid_match,kid_match,knum,kmatrix,label_match)

			block_info={}
			fa={}
			#dlength={}
			dltem={}
			c+=1
		if line[0]=='s':
			ele=line.split()
			seq=ele[-1]
			seq=re.sub('N','',seq)
			pre=' '.join(ele[:-1])
			block_info[ele[1]]=pre
			fa[ele[1]]=seq
			dltem[ele[1]]=len(seq)
			if ele[1] not in dlength:
				dlength[ele[1]]=len(seq)
			else:
				dlength[ele[1]]+=len(seq)
	temc=1
	#query_list=[]
	query_start={}
	ref_start=0
	ref=''
	showp+=1
	res=sorted(dltem.items(), key = lambda kv:(kv[1], kv[0]), reverse = True)
	if not res[0][1]<50:
		#o2=open('other_ref.fasta','w+')
		#query_name='other_ref.fasta'
		for strain in fa:
			if strain not in all_seq:
				all_seq[strain]=[fa[strain]]
			else:
				all_seq[strain].append(fa[strain])
			'''
			if strain==res[0][0]:
				#ref='tem_Ref.fasta'
				#o1=open('tem_Ref.fasta','w+')
				#o1.write('>'+strain+'\n'+fa[strain]+'\n')
				#o1.close()
				if strain not in ref_seq:
					ref_seq[strain]=fa[strain]
				else;
					ref_
				ref_start=int(block_info[strain].split()[2])
			else:
				#query=block_pre+'_tem_'+str(temc)+'.fasta'
				#query_list.append(query)
				query_start[block_info[strain].split()[1]]=int(block_info[strain].split()[2])
				#o2=open(query,'w+')
				o2.write('>'+strain+'\n'+fa[strain]+'\n')
				#o2.close()
				#temc+=1
			'''
	res=sorted(dlength.items(), key = lambda kv:(kv[1], kv[0]), reverse = True)
	#ref='tem_Ref.fasta'
	#query_name='other_ref.fasta'
	#o1=open(ref,'w+')
	#o2=open(query_name,'w+')
	msa_out=os.getcwd()+'/output'
	ref=os.getcwd()+'/ref'
	if not os.path.exists(msa_out):
		os.makedirs(msa_out)
	if not os.path.exists(ref):
		os.makedirs(ref)
	strains=[]
	for strain in all_seq:
		if strain==res[0][0]:
			#strains.append(ref+'/reference.fasta')
			o=open('reference.fasta','w+')
			seq='N'.join(all_seq[strain])
			o.write('>'+strain+'\n'+seq+'\n')
		else:
			strains.append(ref+'/'+strain+'.fasta')
			o=open(ref+'/'+strain+'.fasta','w+')
			seq='N'.join(all_seq[strain])
			o.write('>'+strain+'\n'+seq+'\n')
			
	print('Log: Connet all blocks, done... ')

	align_with_sts(strains,msa_out,ksize,dlabel,sid_match,kid_match,knum,kmatrix,label_match)

	# All pre-work is ready, now to generate SNP matrix and kmer sets
	'''
	print('::Total '+str(len(drq))+' Sites of Ref will be used to construct the matrix')
	print('Now we will generate SNV matrix and kmer sets')
	osnv=open(out_dir+'/'+block_pre+'_matrix.txt','w+')
	ok1=open(out_dir+'/'+block_pre+'_kmer.fasta','w+')
	ok2=open(out_dir+'/'+block_pre+'_kmer.txt','w+')
	# Out k-mers
	kc=1
	ikmer=dict(ikmer)
	for kmr in ikmer:
		ok1.write('>'+str(kc)+'\n'+kmr+'\n')
		kc+=1
		ok2.write(kmr+'\t')
		outs=','.join(ikmer[kmr].keys())
		ok2.write(outs+'\n')
	# Sort all sites
	out_site=sorted(drq.keys())
	out_base=['A','T','G','C']
	osnv.write('\t\t')
	# Output head
	for site in  out_site:
		for base in out_base:
			osnv.write('\t'+str(site)+'-'+base)
	osnv.write('\n')
	for strain in fa:
		osnv.write(strain)
		for site in  out_site:
			for base in out_base:
				if drq[site][strain][1]==base:
					osnv.write('\t1')
				else:
					osnv.write('\t0')
		osnv.write('\n')
	'''



from treelib import Tree, Node
import scipy.stats as st
import os
from collections import defaultdict
import random
import numpy as np
import time
import uuid


def read_tree_structure(db_dir): # tree.data [node category, accessibility, covered_num, total_num, abundance]
    GCF = {}
    f = open(db_dir+"/tree_structure.txt", "r")
    lines = f.readlines()
    tree = Tree()
    sequences = []
    if(lines[-1].split("\t")[1]!="N"):
        for i in range(0, len(lines)):
            if(lines[i].split("\t")[1]=='N'):
                break
        for j in range(i, len(lines)):
            sequences.append(lines[j])
        for j in range(0, i):
            sequences.append(lines[j])
    else:
        sequences = reversed(lines)
    for i in sequences:
        temp = i.rstrip().split("\t")
        if(temp[1] == "N"):
            tree.create_node(identifier=int(temp[0]))
        else:
            tree.create_node(identifier=int(temp[0]), parent=int(temp[1]))
        if(len(temp) == 4):
            GCF[tree.get_node(int(temp[0]))] = temp[3]  # cluster with size 1
    return tree, GCF


def get_node_label(db_dir, tree):
    f = open(db_dir+"/node_length.txt", "r")
    length = {}
    lines = f.readlines()
    for line in lines:
        d = line.rstrip().split("\t")
        length[int(d[0])] = int(d[1])
    for node in tree.all_nodes():
        if(length[node.identifier] < 1000):
            node.data[0] = 0
        elif(length[node.identifier] < 3000):
            node.data[0] = 1
        else:
            node.data[0] = 2
    f = open(db_dir + "/reconstructed_nodes.txt", "r")
    lines = f.readlines()
    for line in lines:
        node = tree.get_node(int(line.rstrip()))
        if(node.data[0] != 0):
            if(length[node.identifier] < 3000):
                node.data[0] = 'o1'
            else:
                node.data[0] = 'o2'


def jellyfish_count(fq_path, db_dir):
    if(type(fq_path) != str):
        fq_path = " ".join(fq_path)
    dir_jf = os.path.split(os.path.abspath(__file__))[0]+'/jellyfish-linux'
    uid = uuid.uuid1().hex
    jf_res_path = "temp_"+uid+".fa"
    os.system(dir_jf + " count -m 31 -s 100M -t 8 --if " + db_dir+"/kmer.fa -o temp_"+uid+".jf " + fq_path)
    os.system(dir_jf+" dump -c temp_"+uid+".jf > temp_"+uid+".fa")
    os.system("rm temp_"+uid+".jf")
    kmer_index_dict = {}
    f = open(db_dir+"/kmer.fa", "r")
    lines = f.readlines()
    for i in range(0, int(len(lines)/2)):
        kmer_index_dict[lines[i*2+1].rstrip()] = i
    f.close()
    match_results = {}
    f = open(jf_res_path, "r")
    lines = f.readlines()
    for line in lines:
        temp = line.rstrip().split(" ")
        match_results[kmer_index_dict[temp[0]]] = int(temp[1])
    os.system("rm temp_"+uid+".fa")
    return match_results


def del_outlier(profile):
    cutoff = 100 * np.median(profile)
    profile_copy = profile.copy()
    for i in profile:
        if(i>=cutoff):
            profile_copy.remove(i)
    return profile_copy


def match_node(match_results, db_dir, node_id, valid_kmers):
    f = open(db_dir+"/kmers/"+str(node_id), "r")
    lines = f.readlines()
    d = set(map(int, lines[0].rstrip().split(" ")))
    valid_kmer = valid_kmers & d
    k_profile = []
    for k in valid_kmer:
        temp = match_results[k]
        if(temp>0): # don't ignore count 1
            k_profile.append(temp)
    if(len(k_profile)>0):
        k_profile = del_outlier(k_profile)
    return len(valid_kmer), k_profile


def piecewise(cov_cutoff, cov, label, k_profile):
    if(label in [1, 'o1']):
        cov_cutoff = cov_cutoff/2
    if(cov >= cov_cutoff):
        return np.mean(k_profile)
    else:
        return 0


def get_uniq_path(node, uniq_path, tree):
    uniq_path.append(node)
    parent = tree.parent(node.identifier)
    if(parent == None or tree.siblings(node.identifier)[0].data[1] in [1, 2]):
        return 1
    get_uniq_path(parent, uniq_path, tree)


def get_ancestor_ab(node, tree, length, cov, abundance):
    uniq_path = []
    get_uniq_path(node, uniq_path, tree)
    kmer_number = {}
    valid_kmers = 0
    for N in uniq_path:
        kmer_number[N] = length[N]*cov[N]
        valid_kmers += length[N]
    kmer_numbers = sum(list(kmer_number.values()))
    if(valid_kmers >= 1000):
        ratio = []
        ab = []
        for N in uniq_path:
            ratio.append(kmer_number[N]/kmer_numbers)
            ab.append(abundance[N])
        return sum([a*b for a,b in zip(ab,ratio)])
    else:
        return -1


def adjust_profile(node, results, valid_kmers, length, abundance, cov, match_results, cov_cutoff, db_dir, overlapping_info):
    overlap = defaultdict(list)
    delete = set([])
    k_profile = []
    f = open(db_dir+"/kmers/"+str(node.identifier), "r")
    lines = f.readlines()
    d = list(map(int, lines[0].rstrip().split(" ")))
    for i in results:
        if(i.identifier in overlapping_info and node.identifier in overlapping_info[i.identifier]):
            overlap[i.identifier] = set([])
            for k in overlapping_info[i.identifier][node.identifier]:
                overlap[i.identifier].add(d[k])
            delete = overlap[i.identifier] | delete
    d = set(d)
    if(len(d) - len(delete) >= 1000):
        remain = d - delete
        valid_kmer = valid_kmers & remain
        for k in valid_kmer:
            temp = match_results[k]
            if(temp>0):
                k_profile.append(temp)
        if(len(k_profile)>0):
            k_profile = del_outlier(k_profile)
        length[node] = len(valid_kmer)
        cov[node] = len(k_profile)/length[node]
        abundance[node] = piecewise(cov_cutoff, cov[node], node.data[0], k_profile)
        if(length[node]<3000):
            return 1
        else:
            return 2
    else:
        valid_kmer = valid_kmers & d
        temp_match = {}
        x = {}
        for i in valid_kmer:
            temp_match[i] = match_results[i]
        for i in results:
            x[i] = i.data[4]
        Tuple = sorted(x.items(), key = lambda kv:(kv[1], kv[0]), reverse = True)
        for j in Tuple:
            temp_match1 = {}
            i = j[0]
            if(i.identifier in overlap):
                overlapped_kmers = valid_kmer&overlap[i.identifier]
                for k in overlapped_kmers:
                    if(temp_match[k] > 0):
                        temp_match1[k] = temp_match[k]
            sample = np.random.poisson(abundance[i], size=len(temp_match1))
            sample.sort()
            Tuple1 = sorted(temp_match1.items(), key = lambda kv:(kv[1], kv[0]))
            for k in range(0, len(sample)):
                temp_match[Tuple1[k][0]] = Tuple1[k][1] - sample[k]
        for i in temp_match:
            if(temp_match[i]>0):
                k_profile.append(temp_match[i])
        length[node] = len(valid_kmer)
        cov[node] = len(k_profile)/length[node]
        abundance[node] = piecewise(cov_cutoff, cov[node], node.data[0], k_profile)
        if(length[node]<3000):
            return 'o1'
        else:
            return 'o2'


def search(pending, match_results, db_dir, valid_kmers, length, cov, abundance, cov_cutoff, ab_cutoff, results, leaves, res_temp, tree, overlapping_info):
    group = pending[0]
    print("__________________________________________________")
    if(len(group)==1 and group[0].data[0]!=0):  # root node is strong
        node = group[0]
        node.data[1] = 1 # access = 1
        length[node], k_profile = match_node(match_results, db_dir, node.identifier, valid_kmers)
        cov[node] = len(k_profile)/length[node]
        abundance[node] = piecewise(cov_cutoff, cov[node], node.data[0], k_profile)
        print("%d:    %f | %f    %d"%(node.identifier, abundance[node], cov[node], length[node]))
        if(abundance[node]>=ab_cutoff):
            pending.append(list(reversed(tree.children(node.identifier))))
        del pending[0]
        return 1
    elif(len(group)==1 and group[0].data[0]==0):    # root node is weak
        node = group[0]
        node.data[1] = 1 # access = 1
        length[node] = 0
        cov[node] = 0
        abundance[node] = 0
        print("%d:    weak"%node.identifier)
        pending.append(list(reversed(tree.children(node.identifier))))
        del pending[0]
        return 1
    print("parent node: %d ->"%tree.parent(group[0].identifier).identifier)
    if(group[0].data[0]==0 and group[0].data[1]==0):    #both weak
        print("%d:    weak\n%d:    weak"%(group[0].identifier, group[1].identifier))
        group[0].data[1] = 2 # access = 2, not sure
        group[1].data[1] = 2
        for node in group:
            abundance[node] = 0
            cov[node] = 0
            length[node] = 0
            pending.append(list(reversed(tree.children(node.identifier))))
        del pending[0]

    correction_label = 0
    group_label = []
    for node in group:
        if(node.data[0] == 0):
            abundance[node] = 0
            cov[node] = 0
            length[node] = 0
            node.data[1] = 2
            pending.append(list(reversed(tree.children(node.identifier))))
            print("%d:    weak"%node.identifier)
            group_label.append((node, 0))
            continue
        elif(node.data[0] in [1, 2] or len(results) == 0):
            if(node.data[0] == 'o1'):
                node.data[0] = 1
            elif(node.data[0] == 'o2'):
                node.data[0] = 2
            group_label.append((node, node.data[0]))
            length[node], k_profile = match_node(match_results, db_dir, node.identifier, valid_kmers)
            if(length[node]<500):
                abundance[node] = 0
                cov[node] = 0
                length[node] = 0
                pending.append(list(reversed(tree.children(node.identifier))))
                print("%d:    weak"%node.identifier)
                group_label.append((node, 0))
            else:
                cov[node] = len(k_profile)/length[node]
                abundance[node] = piecewise(cov_cutoff, cov[node], node.data[0], k_profile)
        else:
            correction_label = 1
            node.data[0] = adjust_profile(node, results, valid_kmers, length, abundance, cov, match_results, cov_cutoff, db_dir, overlapping_info)
            group_label.append((node, node.data[0]))
        if(abundance[node]<ab_cutoff):
            abundance[node] = 0
        print("%d:    %f | %f    %d"%(node.identifier, abundance[node], cov[node], length[node]))

    if(correction_label == 1):
        # calculate reference abundance
        ancestor_ab = get_ancestor_ab(tree.parent(group[0].identifier), tree, length, cov, abundance)
        if(ancestor_ab<=ab_cutoff):
            pass
        else:
            label = 0
            if(set([group_label[0][1], group_label[1][1]]) in [set(['o1', 'o1']), set(['o2', 'o2'])]):
                label = 1
            elif(0 in set([group_label[0][1], group_label[1][1]]) or set([group_label[0][1], group_label[1][1]]) == set(['o1', 'o2'])):
                label = 2
                for i in group_label:
                    if(i[1] == 0 or i[1] == 'o1'):
                        x = i[0]
                    else:
                        y = i[0]
            elif(set([group_label[0][1], group_label[1][1]]) in [set(['o1', 2]), set(['o2', 2])]):
                label = 2
                for i in group_label:
                    if(i[1] == 2):
                        y = i[0]
                    else:
                        x = i[0]
            if(label == 1):
                for i in [group_label[0][0], group_label[1][0]]:
                    abundance[i] = ancestor_ab * (abundance[i]/(abundance[group_label[0][0]]+abundance[group_label[1][0]]))
            elif(label == 2):
                abundance[x] = ancestor_ab - abundance[y]

    # binomial test
    ab_temp = {}
    for i in range(0, 2):
        ab_temp[group[i]] = round(abundance[group[i]])
    if(list(ab_temp.values()) == [0, 0]):
        del pending[0]
        return 1
    Tuple = sorted(ab_temp.items(), key = lambda kv:(kv[1]))
    (a, b, x, y) = (Tuple[1][0], Tuple[0][0], Tuple[1][1], Tuple[0][1])
    ret = 1 - st.binom.sf(max([x, y]), x+y, 0.995)
    if(ret < 0.05):
        temp = (a, b)
    else:
        temp = [a]
    for i in temp:
        if(i.data[0] == 0):
            i.data[1] = 2
        else:
            i.data[1] = 1
        if(i not in leaves):
            if(tree.children(i.identifier) not in pending):
                pending.append(list(reversed(tree.children(i.identifier))))
        else:
            res_temp.append(i)
    del pending[0]
    return 1


def res_node_proc(node, wa_cov_cutoff, length, cov, abundance, tree):
    uniq_path = []
    get_uniq_path(node, uniq_path, tree)
    for j in uniq_path:
        node.data[2] += length[j]*cov[j]
        node.data[3] += length[j]
    node.data[2] = int(node.data[2])
    if(node.data[2]/node.data[3] < wa_cov_cutoff):
        return 0
    ratio = []
    ab = []
    for j in uniq_path:
        ratio.append(cov[j]*length[j]/node.data[2])
        ab.append(abundance[j])
    node.data[4] = sum([a*b for a,b in zip(ab,ratio)])
    if(node.data[4] <= 1):
        return 0
    return 1


def check_access(node, tree):
    node.data[1] = 1
    p = tree.parent(node.identifier)
    if(p != None):
        check_access(p, tree)


def identify_cluster(fq_path, db_dir, cutoff):
    start=time.time()

    tree, GCF = read_tree_structure(db_dir)
    for i in tree.all_nodes():
        i.data = [-1, -1, -1, -1, -1]
    get_node_label(db_dir, tree)
    match_results = jellyfish_count(fq_path, db_dir)
    valid_kmers = set(match_results.keys())
    leaves = tree.leaves()

    # CST search
    ab_cutoff = cutoff[2]
    cov_cutoff = cutoff[0]
    wa_cov_cutoff = cutoff[1]   # weighted average
    pending = [[tree.all_nodes()[0]]]
    results = []
    length = {}
    cov = {}
    abundance = {}
    alternative = []
    overlapping_info = defaultdict(dict)

    while(len(pending)!=0):
        res_temp = []
        search(pending, match_results, db_dir, valid_kmers, length, cov, abundance, cov_cutoff, ab_cutoff, results, leaves, res_temp, tree, overlapping_info)
        for j in res_temp:
            label = res_node_proc(j, wa_cov_cutoff, length, cov, abundance, tree)
            alternative.append(j)
            if(label == 1):
                check_access(j, tree)
                results.append(j)
                if(os.path.exists(db_dir+"/overlapping_info/"+str(j.identifier))):
                    f1 = open(db_dir+"/overlapping_info/"+str(j.identifier), "r")
                    f2 = open(db_dir+"/overlapping_info/"+str(j.identifier)+"_supple", "r")
                    lines = f1.readlines()
                    lines2 = f2.readlines()
                    for line in lines2:
                        d = line.rstrip().split(" ")
                        overlapping_info[j.identifier][int(d[0])] = map(int, lines[int(d[1])].rstrip().split(" "))
            else:
                j.data[1] = 0

    # output
    for i in tree.all_nodes():
        i.data[1] = 0
    for i in results:
        check_access(i, tree)
        i.data[2] = 0
        i.data[3] = 0
    for j in results:
        res_node_proc(j, wa_cov_cutoff, length, cov, abundance, tree)
    total_ab = 0
    if(len(results) > 0):
        for i in results:
            total_ab += i.data[4]
    elif(len(alternative)!=0):
        results = []
        cov_list = {}
        for j in alternative:
            cov_list[j] = j.data[2]/j.data[3]
        r = max(cov_list, key=cov_list.get)
        if(cov_list[r]>=0.1):
            check_access(r, tree)
            label = res_node_proc(j, 0.1, length, cov, abundance, tree)
            if(label == 1):
                results = [r]
                total_ab = r.data[4]
    res = defaultdict(lambda:{})
    for i in results:
        res[i.identifier]['cls_ab'] = i.data[4]
        res[i.identifier]['cls_per'] = i.data[4]/total_ab
        res[i.identifier]['cls_cov'] = i.data[2]/i.data[3]
        res[i.identifier]['cls_total_num'] = i.data[3]
        res[i.identifier]['cls_covered_num'] = i.data[2]
        res[i.identifier]['strain'] = 0
        res[i.identifier]['s_ab'] = 0
        if(i in GCF):
            res[i.identifier]['strain'] = GCF[i]
            res[i.identifier]['s_ab'] = i.data[4]

    end = time.time()
    print('- The total running time of tree search is ',str(end-start),' s\n')
    return res



import re
import os
import psutil
from collections import defaultdict
from Bio import SeqIO
import seqpy
import scipy.sparse as sp
import numpy as np
import multiprocessing
import time
import pickle

def load_cls(cls95):
	d2=defaultdict(lambda:{}) # id -> {strains}
	d3={} # strain -> id
	f2=open(cls95,'r')
	#non_rep={}
	while True:
		line=f2.readline().strip()
		if not line:break
		ele=line.split('\t')
		strains=re.split(',',ele[-1])
		for s in strains:
			d2[int(ele[0])][s]=''
			d3[s]=int(ele[0])
	return d2,d3
	

def build_kmer_dict(ddir,k,dsi,cls_num):
	match_1=[]
	match_2=[]
	for i in range(cls_num):
		match_1.append(i+1)
		match_2.append('0')
	dlabel=defaultdict(lambda:{})
	#dseq=defaultdict(lambda:[]) # id -> [seq]
	#print(dsi)
	#exit()
	for pre in ddir:
		g=ddir[pre]
		seq_dict = {rec.id : rec.seq for rec in SeqIO.parse(g, "fasta")}
		for cl in seq_dict:
			seq=str(seq_dict[cl])
			#dseq[pre].append(seq)
			for i in range(len(seq)-k+1):
				kmer=seq[i:i+k]
				rev_kmer=seqpy.revcomp(seq[i:i+k])
				dlabel[kmer][dsi[pre]-1]=1
				dlabel[rev_kmer][dsi[pre]-1]=1
	return dlabel




def load_strains(infa_strains):
	ddir={} # pre -> strain fasta dir
	for filename in os.listdir(infa_strains):
		pre=re.split('\.',filename)[0]
		g=infa_strains+'/'+filename
		ddir[pre]=g
	return ddir

def build_cls_kmer(infa_strains,cls95):
	k=31
	did2s,ds2id=load_cls(cls95) # ID -> Strains {s1:'',s2:''}/ strain -> ID
	cls_num=len(did2s)
	ddir=load_strains(infa_strains)
	dlabel=build_kmer_dict(ddir,k,ds2id,cls_num) # kmer -> {'c1':'','c2':'',...}
	return dlabel,did2s,cls_num


#def build_parallel(filename,did2s,cls_dir,cls_num,dlabel)
def build_parallel(arg):
	filename=arg[0]
	did2s=arg[1]
	cls_dir=arg[2]
	cls_num=arg[3]
	dlabel=arg[4]

	cid=re.sub('C','',filename)
	print(str(time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time())))+' - StrainScan::build_DB:: C'+str(cid)+'- Build overlap matrix...',flush=True)
	cid=int(cid)
	if len(did2s[cid])>1:
		kid=pickle.load(open(cls_dir+'/'+filename+'/all_kid.pkl','rb'))
		row=len(kid)
		column=len(range(cls_num))
		for k in kid:
			kid[k]=int(kid[k])
		mat=sp.dok_matrix((row,column), dtype=np.int8)
		for k in kid:
			mat[kid[k]-1,list(dlabel[k].keys())]=1
		mat=mat.tocsr()
		sp.save_npz(cls_dir+'/'+filename+'/overlap_matrix.npz',mat)
	print(str(time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time())))+' - StrainScan::build_DB:: C'+str(cid)+u'- Current Memory Usage: %.4f GB' % (psutil.Process(os.getpid()).memory_info().rss / 1024 / 1024 / 1024) )
	print(str(time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time())))+' - StrainScan::build_DB:: C'+str(cid)+'- Omatrix finished',flush=True)
	return

	
	

def build_omatrix(infa_strains,cls95,cls_dir,threads):
	print(str(time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time())))+' - StrainScan::build_DB:: - Build omatrix cls k-mer dict...',flush=True)
	dlabel,did2s,cls_num=build_cls_kmer(infa_strains,cls95)
	print(str(time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time())))+' - StrainScan::build_DB:: - Build omatrix cls k-mer dict... Done',flush=True)
	#para=[]
	dsize={}
	#pool=multiprocessing.Pool(processes=int(threads))
	for filename in os.listdir(cls_dir):
		cid=re.sub('C','',filename)
		cid=int(cid)
		if len(did2s[cid])==1:continue
		dsize[filename]=len(did2s[cid])
	res=sorted(dsize.items(), key = lambda kv:(kv[1], kv[0]), reverse = True)
	for r in res:
		filename=r[0]
		#tem=[filename,did2s,cls_dir,cls_num,dlabel]
		#para.append(tem)
		build_parallel([filename,did2s,cls_dir,cls_num,dlabel])
	#pool.close()
	#pool.join()
	print(str(time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time())))+' - StrainScan::build_DB:: - All omatrices are finished now.',flush=True)
	return
	#os.system('rm -rf '+rdir)
	

#build_omatrix('../Cutibacterium_acnes_small_pure','../Csmall_50/Cluster_Result/hclsMap_95.txt','../Csmall_50/Kmer_Sets_L2/Kmer_Sets')
#build_omatrix('../../Ref_Genome','../DB_Cae_20/Cluster_Result/hclsMap_95.txt','../DB_Cae_20/Kmer_Sets_L2/Kmer_Sets')

	



import re
import os
import sys
import argparse
sys.path.append('library')
from library import Cluster,Unique_kmer_detect_direct,Build_kmer_sets_unique_region_lasso_test_allinone_sp,select_rep,Build_overlap_matrix_sp,Build_tree
import time
import psutil
from collections import defaultdict



__author__="Liao Herui, Ji Yongxin - PhD of City University of HongKong"
usage="StrainScan - A kmer-based strain-level identification tool."

def initial_para(para,value):
	if not para:
		para=value
	return para

def build_dir(in_dir):
	if not os.path.exists(in_dir):
		os.makedirs(in_dir)
def merge_cls(dc_in):
	dc_out=defaultdict(lambda:{})
	for e in dc_in:
		for e2 in dc_in[e]:
			dc_out['C'][e2]=''
	return dict(dc_out)


def mannual(icf,fa_dir):
	dn={} # pre -> full file dir
	for filename in os.listdir(fa_dir):
		pre=re.split('\.',filename)[0]
		dn[pre]=fa_dir+'/'+filename
	f=open(icf,'r')
	dc95_l2=defaultdict(lambda:{})
	while True:
		line=f.readline().strip()
		if not line:break
		ele=line.split('\t')
		st=re.split(',',ele[-1])
		for s in st:
			dc95_l2[int(ele[0])][dn[s]]=''
	return dc95_l2

	

def main():
	pwd=os.getcwd()
	# Get para
	parser=argparse.ArgumentParser(prog='StrainScan_build.py',description=usage)
	parser.add_argument('-i','--input_fasta',dest='input_fa',type=str,required=True,help="The dir of input fasta genome --- Required")
	parser.add_argument('-o','--output_dir',dest='out_dir',type=str,help='Output dir (default: current dir/StrainVote_DB)')
	parser.add_argument('-k','--kmer_size',dest='ksize',type=str,help='The size of kmer, should be odd number. (default: k=31)')
	parser.add_argument('-t','--threads',dest='threads',type=str,help='The threads used to build the database. (default: t=1)')
	parser.add_argument('-u','--uk_num',dest='uknum',type=str,help='The maximum number of unique k-mers in each genome to extract. (default: u=100000)')
	parser.add_argument('-g','--gk_ratio',dest='gkratio',type=str,help='The ratio of group-specific k-mers to extract. (default: g=1.0)')
	parser.add_argument('-m','--strainest_sample',dest='mas',type=str,help='If this parameter is 1, then the program will search joint kmer sets from msa generated by Strainest. (default: 0)')
	parser.add_argument('-n','--mink_cutoff',dest='mink',type=str,help='Minimum k-mer number cutoff in a node (default: n=1000)')
	parser.add_argument('-x','--maxk_cutoff',dest='maxk',type=str,help='Maximum k-mer number cutoff in a node (default: x=30000)')
	parser.add_argument('-r','--maxn_cutoff',dest='maxn',type=str,help='Maximum cluster number for node reconstruction (default: r=3000)')

	#parser.add_argument('-l','--input_clster_file',dest='icf',type=str,help='The cluster file offered by users for mannual database construction')

	args=parser.parse_args()
	fa_dir=args.input_fa
	out_dir=args.out_dir
	ksize=args.ksize
	if not args.threads:
		threads='1'
	else:
		threads=str(args.threads)
	if not args.uknum:
		uknum=100000
	else:
		uknum=int(args.uknum)
	if not args.gkratio:
		gkratio=1.0
	else:
		gkratio=float(args.gkratio)
	if not args.mas:
		mas=0
	else:
		mas=int(args.mas)
	if not args.mink:
		mink=1000
	else:
		mink=int(args.mink)
	if not args.maxk:
		maxk=30000
	else:
		maxk=int(args.maxk)
	if not args.maxn:
		maxn=3000
	else:
		maxn=int(args.maxn)
	params=[0.8,mink,maxk,maxn]
	
	ksize=initial_para(ksize,31)

	#icf=args.icf
	'''
	tid=args.tid
	if not tid:
		tid='1747'
	'''

	out_dir=initial_para(out_dir,pwd+'/StrainScan_DB')
	if not re.search('/',out_dir):
		out_dir=pwd+'/'+out_dir
	cls_res=out_dir+'/Cluster_Result'
	#tree_dir=out_dir+'/Tree_database'
	#kk_db=out_dir+'/Krakenuniq_DB'
	#kk_db1=out_dir+'/Krakenuniq_DB/library'
	#kk_db2=out_dir+'/Krakenuniq_DB/taxonomy'
	#uni_kmer=out_dir+'/Unique_Kmer'
	#union_kmer=out_dir+'/Union_Kmer'
	#kmer_sets_l1=out_dir+'/Kmer_Sets_L1'
	kmer_sets_l2=out_dir+'/Kmer_Sets_L2'
	#tem_dir=out_dir+'/Tem_dir'
	

	
	build_dir(out_dir)
	build_dir(cls_res)
	#build_dir(kk_db)
	#build_dir(kk_db1)
	#build_dir(kk_db2)
	#os.system('cp library/taxonomy/* '+kk_db2)
	#build_dir(uni_kmer)
	#build_dir(tree_dir)
	#build_dir(uni_kmer)
	#build_dir(kmer_sets_l1)
	build_dir(kmer_sets_l2)
	#build_dir(tem_dir)
	
	# Construct matrix with dashing (jaccard index)
	matrix=Cluster.construct_matrix(fa_dir)
	
	# -------- Hirarchical clustering Part --------
	#### Default: Single: 0.95, Complete: 0.95
	# ---------------------------------------------
	dc95=Cluster.hcls(matrix,'single','0.05')
	os.system('mv hclsMap_* distance_matrix_rebuild.txt distance_matrix.txt '+cls_res)
	cls_file=cls_res+'/hclsMap_95.txt'
	dc95_rep,dc95_l2=select_rep.pick_rep(cls_res+'/distance_matrix_rebuild.txt',cls_file,cls_res)

	# Extract k-mers from all strains
	#dlabel,did2k=extract_kmers_from_ref_v3.extract_kmer(fa_dir,ksize,cls_file)
	#exit()
	

	# Construct the tree 	
		
	build_dir(out_dir+'/Tree_database/test')
	build_dir(out_dir+'/Tree_database/nodes_kmer')
	build_dir(out_dir+'/Tree_database/overlap')
	Build_tree.build_tree([cls_res+'/distance_matrix.txt',cls_res+'/hclsMap_95_recls.txt',out_dir+'/Tree_database',31,params])
	
	icf=out_dir+'/Tree_database/hclsMap_95_recls.txt'
	os.system('cp '+out_dir+'/Tree_database/hclsMap_95_recls.txt '+cls_res)
	dc95_l2=mannual(icf,fa_dir)
	
	# Delete tem dir
	os.system('rm -rf '+out_dir+'/Tree_database/test')


	# --------- Inside cluster strains kmer sets construction -------
	Build_kmer_sets_unique_region_lasso_test_allinone_sp.build_kmer_sets(dc95_l2,kmer_sets_l2,ksize,uknum,gkratio,mas,threads)
	print(str(time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time())))+' - StrainScan::build_DB:: Build sparse matices done '+u'- Current Memory Usage: %.4f GB' % (psutil.Process(os.getpid()).memory_info().rss / 1024 / 1024 / 1024) )
	#exit()
	
	#exit()
	# --------- Build Overlap matrix -----------
	new_cls_file=out_dir+'/Tree_database/hclsMap_95_recls.txt'
	Build_overlap_matrix_sp.build_omatrix(fa_dir,new_cls_file,kmer_sets_l2+'/Kmer_Sets',threads)

	# --------- Delete tem dir ---------
	os.system('rm -rf '+kmer_sets_l2+'/Colinear_Block')
	


if __name__=='__main__':
	sys.exit(main())



import setuptools

setuptools.setup(
    name="strainscan",
    version="1.0.4",
    author="Liao Herui and Ji Yongxin",
    author_email="heruiliao2-c@my.cityu.edu.hk",
    description="One efficient and accurate strain-level microbiome composition analysis tool based on reference genomes and k-mers.",
    long_description="StrainScan takes reference database and sequencing data as input, outputs strain-level microbiome compistion analysis report.",
    long_description_content_type="text/markdown",
    url="https://github.com/liaoherui/StrainScan",
    packages=setuptools.find_packages(),
    include_package_data=True,
    zip_safe=True,
    package_data={"StrainScan":["library/jellyfish-linux","library/hier.R","library/dashing_s128","library/seqpy.c","library/seqpy.cpython-37m-x86_64-linux-gnu.so"]},
    install_requires=[
    "numpy==1.17.3",
    "pandas==1.0.1",
    "biopython==1.74",
    "scipy==1.3.1",
    "scikit-learn==0.23.1",
    "bidict==0.21.3",
    "treelib==1.6.1"
    ],
    entry_points={
        'console_scripts':[
        "strainscan = StrainScan.StrainScan:main",
        "strainscan_build = StrainScan.StrainScan_build:main",
        ]
    },
    python_requires='~=3.7',
    classifiers=[
    'Programming Language :: Python :: 3.7',
    ],
)



from scipy.spatial.distance import pdist,squareform
#from scipy.cluster.hierarchy import linkage, dendrogram,fcluster
import os
import re
import numpy as np
import pandas as pd
import pickle
import scipy.sparse as sp
#a=np.array(['1','0','0','1','1','1','0'])
#b=np.array(['0','0','1','1','1','1','1'])
def cal_dist(u,v):
	l=np.count_nonzero(u!=v)
	return l
	#print(l)

def remove_1per(in_csv,idp,out):
	#data=pd.read_csv("all_strain.csv")
	#data=pd.read_csv(in_csv)
	data=sp.load_npz(in_csv)
	data=data.A
	#X=data.to_numpy()
	X=data.T

	total_kmer=np.sum(X,axis=1)
	total_kmer=np.array(total_kmer)
	total_kmer[total_kmer==0]=1

	#total_kmer=np.sum(X,axis=1)
	dm=squareform(pdist(X,cal_dist))
	distance_matrix=dm/total_kmer[:,None]

	sid_match=pickle.load(open(idp, "rb"))
	sk=dict(zip(sid_match,list(total_kmer)))# Dict : strain -> total kmer

	temd=pd.DataFrame(distance_matrix,index=sid_match,columns=sid_match)
	temd.to_csv(out+'/tem_dist.csv',sep="\t")
	ot=open(out+'/tem_hier.R','w+')
	ot.write('x<-read.table(\"'+out+'/tem_dist.csv\", header=T, row.names=1)\nd<-as.dist(as(x,\"matrix\"))\nhc<-hclust(d,method=\"complete\")\nres<-sort(cutree(hc,h=0.01))\nres') # Cutoff: 99.9% or 99%
	ot.close()
	os.system('Rscript '+out+'/tem_hier.R > '+out+'/cls_res.txt')
	os.system('rm '+out+'/tem_hier.R '+out+'/tem_dist.csv')
	f=open(out+'/cls_res.txt','r')
	a=[]
	while True:
		line=f.readline().strip()
		if not line:break
		a.append(line)
	d={}	
	#dmap={}
	c=0
	for l in a[::-1]:
		c+=1
		if not c%2==0:
			ele=l.split()
			if len(ele)==1:
				if l not in d:
					d[int(l)]={}
					#dmap[l]={}
				name=int(l)
			else:
				for e in ele:
					if int(e) not in d:
						d[int(e)]={}
				name=ele
		else:
			ele=l.split()
			if len(ele)==1:
				d[name][l]=''
			else:
				i=0
				for e in ele:
					d[int(name[i])][e]=''
					i+=1
	f.close()
	os.system('rm '+out+'/cls_res.txt')	

	nsid_match={}

	ni=1
	for s in sid_match:
		nsid_match[s]=str(ni)
		ni+=1


	def pick_rep(in_cls,sk):
		max_kmer=0
		rep=''
		for s in in_cls:
			if sk[s]>max_kmer:
				max_kmer=sk[s]
				rep=s
		return rep

	o1=open(out+'/Re_Cluster_info.txt','w+')
	left=[]
	remain=[]
	strains=[]
	#print(sorted(d.keys()))
	#exit()
	for cid in sorted(d.keys()):
		rep=pick_rep(d[cid],sk)
		#print(cid,rep,nsid_match[rep])
		left.append(nsid_match[rep])
		remain.append(int(nsid_match[rep])-1)
		strains.append(rep)
		o1.write(str(cid)+'\t'+rep+'\t'+str(sk[rep])+'\t'+str(len(d[cid]))+'\t'+','.join(list(d[cid].keys()))+'\n')
	#ndata=data.loc[:, left]
	#ndata.to_csv(out+'/all_strains_re.csv',index=False)
	#print(remain)
	ndata=data[:,remain]
	ndata=sp.csr_matrix(ndata)
	sp.save_npz(out+'/all_strains_re.npz',ndata)

	with open(out+'/id2strain_re.pkl','wb') as o2:
		pickle.dump(strains, o2, pickle.HIGHEST_PROTOCOL)

